import copy

import pandas as pd

from immuneML.ml_methods.classifiers.MLMethod import MLMethod
from immuneML.util.Logger import print_log
from immuneML.workflows.steps.MLMethodTrainerParams import MLMethodTrainerParams
from immuneML.workflows.steps.Step import Step


class MLMethodTrainer(Step):

    @staticmethod
    def run(input_params: MLMethodTrainerParams = None):

        print_log(f"ML model training started...", include_datetime=True)

        method = MLMethodTrainer._fit_method(input_params)
        MLMethodTrainer.store(method, input_params)

        print_log(f"ML model training finished.", include_datetime=True)

        return method

    @staticmethod
    def _fit_method(input_params: MLMethodTrainerParams):
        method = copy.deepcopy(input_params.method)
        method.result_path = input_params.result_path

        if input_params.model_selection_cv:
            method.fit_by_cross_validation(encoded_data=input_params.dataset.encoded_data,
                                           number_of_splits=input_params.model_selection_n_folds,
                                           label=input_params.label,
                                           cores_for_training=input_params.cores_for_training,
                                           optimization_metric=input_params.optimization_metric)
        else:
            method.fit(encoded_data=input_params.dataset.encoded_data,
                       label=input_params.label,
                       cores_for_training=input_params.cores_for_training,
                       optimization_metric=input_params.optimization_metric)

        return method

    @staticmethod
    def store(method: MLMethod, input_params: MLMethodTrainerParams):
        method.store(input_params.result_path)
        train_predictions = method.predict(input_params.dataset.encoded_data, input_params.label)
        train_proba_predictions = method.predict_proba(input_params.dataset.encoded_data, input_params.label)

        df = pd.DataFrame({"example_ids": input_params.dataset.encoded_data.example_ids,
                           f"{input_params.label.name}_predicted_class": train_predictions[input_params.label.name],
                           f"{input_params.label.name}_true_class": input_params.dataset.encoded_data.labels[input_params.label.name]})

        for cls in method.get_classes():
            tmp = train_proba_predictions[input_params.label.name][cls] if train_proba_predictions is not None and train_proba_predictions[input_params.label.name] is not None else None
            df[f'{input_params.label.name}_{cls}_proba'] = tmp

        df.to_csv(input_params.train_predictions_path, index=False)



import logging

from immuneML.data_model.SequenceSet import ReceptorSequence
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.encodings.kmer_frequency.sequence_encoding.SequenceEncodingStrategy import SequenceEncodingStrategy
from immuneML.environment.Constants import Constants
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.util.KmerHelper import KmerHelper


class IMGTKmerSequenceEncoder(SequenceEncodingStrategy):

    @staticmethod
    def encode_sequence(sequence: ReceptorSequence, params: EncoderParams):
        """
        creates overlapping continuous k-mers from a sequence as features for use in KmerFrequencyEncoder
        object of type EncoderParams, same object as passed into KmerFrequencyEncoder
        :param sequence: ReceptorSequence
        :param params: EncoderParams (where params["model"]["k"] is used)
        :return: SequenceEncodingResult consisting of features and feature information names
        """
        k = params.model["k"]
        sequence_type = params.model.get('sequence_type', params.sequence_type)
        length = len(sequence.get_sequence(sequence_type))

        if length < k:
            logging.warning('Sequence length is less than k. Ignoring sequence')
            return None

        kmers = KmerHelper.create_IMGT_kmers_from_sequence(sequence=sequence, k=k, sequence_type=sequence_type,
                                                           region_type=params.model.get('region_type',
                                                                                        params.region_type))

        kmers = [Constants.FEATURE_DELIMITER.join([str(mer) for mer in kmer]) for kmer in kmers]

        return kmers

    @staticmethod
    def get_feature_names(params: EncoderParams):
        return ["sequence", "imgt_position"]



import logging
from pathlib import Path

import pandas as pd
import plotly.express as px

from immuneML.data_model.EncodedData import EncodedData
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.dsl.definition_parsers.MLParser import MLParser
from immuneML.ml_methods.dim_reduction.DimRedMethod import DimRedMethod
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.encoding_reports.EncodingReport import EncodingReport
from immuneML.util.PathBuilder import PathBuilder


class DimensionalityReduction(EncodingReport):
    """
    This report visualizes the data obtained by dimensionality reduction.

    **Specification arguments:**

    - label (str): name of the label to use for highlighting data points; or None

    - dim_red_method (str): dimensionality reduction method to be used for plotting; if set, in a workflow, this
      dimensionality reduction will be used for plotting instead of any other set in the workflow; if None, it will
      visualize the encoded data of reduced dimensionality if set


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            reports:
                rep1:
                    DimensionalityReduction:
                        label: epitope
                        dim_red_method:
                            PCA:
                                n_components: 2

    """

    @classmethod
    def build_object(cls, **kwargs):
        if "dim_red_method" in kwargs and kwargs["dim_red_method"] and isinstance(kwargs["dim_red_method"], dict):
            cls_name = list(kwargs['dim_red_method'].keys())[0]
            method = MLParser.parse_any_model("dim_red_method", kwargs['dim_red_method'], cls_name)[0]
        else:
            method = None
        return DimensionalityReduction(**{**kwargs, "dim_red_method": method})

    def __init__(self, dataset: Dataset = None, batch_size: int = 1, result_path: Path = None,
                 name: str = None, label: str = None, dim_red_method: DimRedMethod = None):
        super().__init__(dataset=dataset, result_path=result_path, name=name)
        self._label = label
        self._dim_red_method = dim_red_method
        self._dimension_names = ['dimension_1',
                                 'dimension_2'] if self._dim_red_method else self._dim_red_method.get_dimension_names()
        self.info = "This report visualizes the encoded data after applying dimensionality reduction"

    def check_prerequisites(self):
        return (isinstance(self.dataset.encoded_data, EncodedData) and
                (self.dataset.encoded_data.dimensionality_reduced_data is not None or self._dim_red_method is not None))

    def _generate(self) -> ReportResult:
        if self._dim_red_method:
            assert self.dataset.encoded_data.examples is not None, \
                f"{DimensionalityReduction.__name__}: data not encoded, report will not be made."
            dim_reduced_data = self._dim_red_method.fit_transform(self.dataset)
        else:
            assert self.dataset.encoded_data.dimensionality_reduced_data is not None
            dim_reduced_data = self.dataset.encoded_data.dimensionality_reduced_data

        assert dim_reduced_data.shape[1] == 2
        data_labels = None

        try:
            data_labels = self.dataset.get_attribute(self._label).tolist()
        except (AttributeError, TypeError) as e:
            logging.warning(f"Label {self._label} not found in the dataset. Skipping label coloring in the plot.")

        PathBuilder.build(self.result_path)

        df = pd.DataFrame({'example_id': self.dataset.get_example_ids(),
                           self._dimension_names[0]: dim_reduced_data[:, 0],
                           self._dimension_names[1]: dim_reduced_data[:, 1]})
        if self._label:
            df[self._label] = data_labels
        df.to_csv(self.result_path / 'dimensionality_reduced_data.csv', index=False)

        report_output_fig = self._safe_plot(df=df, output_written=True)
        output_figures = None if report_output_fig is None else [report_output_fig]

        dim_red_text = f" ({self._dim_red_method.__class__.__name__})." if self._dim_red_method else "."

        return ReportResult(name=self.name, info=self.info + dim_red_text,
                            output_figures=output_figures,
                            output_tables=[ReportOutput(self.result_path / 'dimensionality_reduced_data.csv',
                                                        'data after dimensionality reduction')])

    def _plot(self, df: pd.DataFrame) -> ReportOutput:
        if self._label:
            # Convert to categorical if few unique values
            unique_values = df[self._label].unique()
            if len(unique_values) <= 3:
                df[self._label] = df[self._label].astype('category')
                figure = px.scatter(df, x=self._dimension_names[0], y=self._dimension_names[1], color=self._label,
                                    color_discrete_sequence=px.colors.qualitative.Set1,
                                    category_orders={self._label: sorted(unique_values)})
            else:
                figure = px.scatter(df, x=self._dimension_names[0], y=self._dimension_names[1], color=self._label)
        else:
            # No label case - just plot points
            figure = px.scatter(df, x=self._dimension_names[0], y=self._dimension_names[1])

        figure.update_layout(template="plotly_white")

        PathBuilder.build(self.result_path)

        file_path = self.result_path / "dimensionality_reduction.html"
        figure.write_html(str(file_path))
        return ReportOutput(path=file_path, name="Data visualization after dimensionality reduction")



import numpy as np


def jaccard(vector1, vector2, tmp_vector=None):
    return 1 - np.sum(np.logical_and(vector1, vector2, out=tmp_vector)) / np.sum(np.logical_or(vector1, vector2, out=tmp_vector))

def morisita_horn(vector1, vector2, *args, **kwargs):
    sum1 = np.sum(vector1)
    sum2 = np.sum(vector2)

    return 1 - ((2 * np.sum(vector1 * vector2)) / ((np.sum(vector1 ** 2) / (sum1 ** 2) + np.sum(vector2 ** 2) / (sum2**2)) * sum1 * sum2))



import logging
import os
import re
from pathlib import Path

import yaml

from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.util.ReflectionHandler import ReflectionHandler


class DefaultParamsLoader:

    @staticmethod
    def convert_to_snake_case(name):
        if name not in ["MiXCR", "VDJdb", "TCRdistMotifDiscovery", "LIgOSimulation", "SoNNia"]:
            s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
            return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()
        elif name in ["TCRdistMotifDiscovery"]:
            return re.sub('([a-z])([A-Z])', r'\1_\2', name).lower()
        else:
            return name.lower()

    @staticmethod
    def _convert_to_camel_case(name):
        return ''.join([i.title() for i in name.split('_')])

    @staticmethod
    def load(path, class_name, log_if_missing=True):
        if os.path.isabs(path):
            filepath = Path(path) / f"{DefaultParamsLoader.convert_to_snake_case(class_name)}_params.yaml"
        else:
            filepath = EnvironmentSettings.default_params_path / path / f"{DefaultParamsLoader.convert_to_snake_case(class_name)}_params.yaml"

        if filepath.is_file():
            with filepath.open("r") as file:
                params = yaml.load(file, Loader=yaml.FullLoader)
        else:
            if log_if_missing:
                logging.info("DefaultParams: no default parameters were found for {}. Proceeding...".format(class_name))
            params = {}

        return params

    @staticmethod
    def _parse_to_enum_instances(params, location):
        for key in params.keys():
            class_name = DefaultParamsLoader._convert_to_camel_case(key)
            if ReflectionHandler.exists(class_name, location):
                cls = ReflectionHandler.get_class_by_name(class_name, location)
                params[key] = cls[params[key].upper()]
        return params



from abc import ABC, abstractmethod
import numpy as np

from immuneML.caching.CacheHandler import CacheHandler
from immuneML.data_model.AIRRSequenceSet import AIRRSequenceSet
from immuneML.data_model.EncodedData import EncodedData
from immuneML.data_model.SequenceParams import RegionType
from immuneML.data_model.bnp_util import get_sequence_field_name
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.data_model.datasets.ElementDataset import SequenceDataset, ReceptorDataset
from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.encodings.DatasetEncoder import DatasetEncoder
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.environment.SequenceType import SequenceType
from immuneML.util.ParameterValidator import ParameterValidator


class ProteinEmbeddingEncoder(DatasetEncoder, ABC):
    """
    Abstract base class for protein embedding encoders that handles dataset-type specific logic.
    Subclasses must implement the _embed_sequence_set method.
    """

    def __init__(self, region_type: RegionType, name: str = None):
        super().__init__(name)
        self.region_type = region_type

    @staticmethod
    @abstractmethod
    def build_object(dataset: Dataset, **params):
        pass

    def encode(self, dataset: Dataset, params: EncoderParams) -> Dataset:
        cache_params = self._get_caching_params(dataset, params)
        if isinstance(dataset, SequenceDataset):
            return CacheHandler.memo_by_params(cache_params, lambda: self._encode_sequence_dataset(dataset, params))
        elif isinstance(dataset, ReceptorDataset):
            return CacheHandler.memo_by_params(cache_params, lambda: self._encode_receptor_dataset(dataset, params))
        elif isinstance(dataset, RepertoireDataset):
            return CacheHandler.memo_by_params(cache_params, lambda: self._encode_repertoire_dataset(dataset, params))
        else:
            raise RuntimeError(f"{self.__class__.__name__}: {self.name}: invalid dataset type: {type(dataset)}.")

    def _encode_sequence_dataset(self, dataset: SequenceDataset, params: EncoderParams):
        seq_field = get_sequence_field_name(self.region_type, SequenceType.AMINO_ACID)
        embeddings = self._embed_sequence_set(dataset.data, seq_field)

        encoded_dataset = dataset.clone()
        encoded_dataset.encoded_data = EncodedData(examples=np.array(embeddings),
                                                   labels={label.name: getattr(dataset.data, label.name).tolist()
                                                           for label in params.label_config.get_label_objects()},
                                                   example_ids=dataset.data.sequence_id.tolist(),
                                                   encoding=self._get_encoding_name())
        return encoded_dataset

    def _encode_receptor_dataset(self, dataset: ReceptorDataset, params: EncoderParams):
        seq_field = get_sequence_field_name(region_type=self.region_type, sequence_type=SequenceType.AMINO_ACID)

        data = dataset.data

        loci = sorted(list(set(data.locus.tolist())))
        assert len(loci) == 2, (
            f"{self.__class__.__name__}: {self.name}: to encode receptor dataset, it has to include "
            f"two different chains, but got: {loci} instead.")

        embeddings = self._embed_sequence_set(data, seq_field)

        cell_ids = data.cell_id.tolist()
        chain_types = data.locus.tolist()

        chain1_embeddings = {}
        chain2_embeddings = {}

        for i, (cell_id, chain_type) in enumerate(zip(cell_ids, chain_types)):
            if chain_type == loci[0]:
                chain1_embeddings[cell_id] = embeddings[i]
            else:
                chain2_embeddings[cell_id] = embeddings[i]

        assert set(chain1_embeddings.keys()) == set(chain2_embeddings.keys()), \
            f"{self.__class__.__name__}: {self.name}: some receptors are missing one of the chains"

        receptor_ids = list(chain1_embeddings.keys())
        concatenated_embeddings = np.array([
            np.concatenate([chain1_embeddings[cell_id], chain2_embeddings[cell_id]])
            for cell_id in receptor_ids
        ])

        labels = (data.topandas().groupby('cell_id').first()[params.label_config.get_labels_by_name()]
                  .to_dict(orient='list'))

        encoded_dataset = dataset.clone()
        encoded_dataset.encoded_data = EncodedData(
            examples=concatenated_embeddings,
            labels=labels, example_ids=receptor_ids,
            encoding=self._get_encoding_name()
        )

        return encoded_dataset

    def _encode_repertoire_dataset(self, dataset: RepertoireDataset, params: EncoderParams) -> Dataset:
        seq_field = get_sequence_field_name(self.region_type, SequenceType.AMINO_ACID)

        examples = []
        for repertoire in dataset.repertoires:
            examples.append(CacheHandler.memo_by_params((repertoire.identifier, self.__class__.__name__,
                                                         self.region_type.name, self._get_model_link()),
                                                        lambda: self._avg_sequence_set_embedding(
                                                            embedding=self._embed_sequence_set(repertoire.data,
                                                                                               seq_field))))

        encoded_dataset = dataset.clone()
        labels = dataset.get_metadata(params.label_config.get_labels_by_name())
        encoded_dataset.encoded_data = EncodedData(examples=np.array(examples), labels=labels,
                                                   example_ids=dataset.get_example_ids(),
                                                   encoding=self._get_encoding_name())

        return encoded_dataset

    def _avg_sequence_set_embedding(self, embedding: np.ndarray) -> np.ndarray:
        return embedding.mean(axis=1)

    @abstractmethod
    def _embed_sequence_set(self, sequence_set: AIRRSequenceSet, seq_field: str) -> np.ndarray:
        pass

    @abstractmethod
    def _get_encoding_name(self) -> str:
        pass

    @abstractmethod
    def _get_model_link(self) -> str:
        pass

    @abstractmethod
    def _get_caching_params(self, dataset, params: EncoderParams):
        pass



import dataclasses
from numbers import Number

import numpy as np
import scipy.special
from bionumpy.encoded_array import Encoding, as_encoded_array, EncodedRaggedArray, EncodedArray
from bionumpy.encodings import AlphabetEncoding, AminoAcidEncoding
from npstructures import RaggedArray, RaggedShape


class EncodedLookup(np.lib.mixins.NDArrayOperatorsMixin):
    def __init__(self, lookup: np.ndarray, encoding: Encoding):
        self._lookup = np.asanyarray(lookup)
        self._encoding = encoding

    def __repr__(self):
        return repr(self._lookup)[:20]

    @property
    def alphabet_size(self):
        return self._lookup.shape[-1]

    @property
    def encoding(self):
        return self._encoding

    def raw(self):
        return self._lookup

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        assert method == "__call__"
        if isinstance(inputs, tuple):
            inputs = tuple(i._lookup if isinstance(i, EncodedLookup) else i for i in inputs)
        else:
            inputs = inputs._lookup if isinstance(inputs, EncodedLookup) else inputs
        return self.__class__(ufunc(*inputs, **kwargs), self._encoding)

    def __getitem__(self, key):
        assert not isinstance(key, Number), key
        key, shape = self._translate_key(key)
        value = self._lookup[key]
        if shape:
            value = RaggedArray(value, shape[-1])
        return value

    def __setitem__(self, key, value):
        key, shape = self._translate_key(key)
        self._lookup[key] = value

    def _translate_key(self, key):
        shape = None
        if isinstance(key, tuple):
            key = tuple(as_encoded_array(i, self._encoding).raw() if (i is not Ellipsis and i is not None) else i
                        for i in key)
            shape = tuple(i.shape if isinstance(i, RaggedArray) else (1,) for i in key)
            key = tuple(i.ravel() if isinstance(i, RaggedArray) else i for i in key)
        else:
            key = as_encoded_array(key, self._encoding).raw()
            if isinstance(key, RaggedArray):
                shape = key.shape
                key = key.ravel()
        return key, shape


@dataclasses.dataclass
class SequenceTransitionDistribution:
    transition_matrix: EncodedLookup
    initial_distribution: EncodedLookup
    end_probs: EncodedLookup

    @classmethod
    def from_probabilities(cls, *args, **kwargs):
        return cls(*(np.log(arg) for arg in args), **{k: np.log(v) for k, v in kwargs.items()})

    def log_prob(self, sequence):
        sequence = as_encoded_array(sequence, self.transition_matrix.encoding)
        prev = sequence[..., :-1]
        next = sequence[..., 1:]
        transition_log_probs = self.transition_matrix[(prev, next)].sum(axis=-1)
        start_log_probs = self.initial_distribution[sequence[..., 0]]
        end_log_probs = self.end_probs[sequence[..., -1]]
        return transition_log_probs + start_log_probs + end_log_probs

    def save(self, filename):
        np.savez(filename, transition_matrix=self.transition_matrix._lookup,
                 initial_distribution=self.initial_distribution._lookup,
                 end_probs=self.end_probs._lookup)

    def sample(self, n_samples):
        encoding = self.transition_matrix.encoding
        n_letters = encoding.alphabet_size
        transitions = np.exp(self.transition_matrix.raw())
        initial_probs = np.exp(self.initial_distribution.raw())
        end_probs = np.exp(self.end_probs.raw())
        matrix = np.hstack((transitions, end_probs[:, None]))
        matrix = np.concatenate([matrix, [np.append(initial_probs, 0)]])
        state = -1
        choices = np.arange(n_letters+1)
        data = []
        lens = []
        len_counter = 0
        while len(lens) < n_samples:
            assert np.allclose(matrix[state].sum(), 1), matrix[state]
            state = np.random.choice(choices, p=matrix[state])
            if state == n_letters:
                lens.append(len_counter)
                len_counter = 0
                continue
            len_counter += 1
            data.append(state)
        assert len(lens) == n_samples, (len(lens), n_samples)
        return EncodedRaggedArray(
            EncodedArray(np.array(data, dtype=int), encoding),
            lens)

    @classmethod
    def load(cls, filename):
        encoding = AminoAcidEncoding
        data = np.load(filename)
        return cls(EncodedLookup(data['transition_matrix'], encoding),
                   EncodedLookup(data['initial_distribution'], encoding),
                   EncodedLookup(data['end_probs'], encoding))


def estimate_transition_model(sequences, weights=None):
    pseudo_count = 1
    n = len(sequences)
    encoding: AlphabetEncoding = sequences.encoding
    n_letters = encoding.alphabet_size
    transition_pseudo = pseudo_count*np.mean(sequences.lengths)/(n_letters + 1)/n_letters
    initial_distribution = np.bincount(sequences[..., 0].raw(), weights=weights, minlength=n_letters) + pseudo_count/n_letters
    end_probs = np.bincount(sequences[..., -1].raw(), weights=weights, minlength=n_letters) + transition_pseudo
    # denominator = np.bincount(sequences.raw().ravel(), minlength=n_letters) + pseudo_count
    if weights is not None:
        weights = RaggedShape(sequences.lengths - 1).broadcast_values(weights[:, None]).ravel()
    transition_matrix = np.bincount(
        np.ravel_multi_index((sequences[..., :-1].raw().ravel(), sequences[..., 1:].raw().ravel()),
                             (n_letters, n_letters)),
        weights=weights, minlength=n_letters ** 2).reshape((n_letters, n_letters)) + transition_pseudo
    row_sums = transition_matrix.sum(axis=1)+end_probs
    transition_matrix = transition_matrix / row_sums[..., None]
    initial_distribution = initial_distribution / initial_distribution.sum()
    end_probs = end_probs / row_sums
    prob_sum = np.sum(transition_matrix, axis=-1) + end_probs
    assert np.allclose(prob_sum, 1), (np.sum(transition_matrix, axis=-1), end_probs)

    return SequenceTransitionDistribution.from_probabilities(
        EncodedLookup(transition_matrix, encoding),
        EncodedLookup(initial_distribution, encoding),
        EncodedLookup(end_probs, encoding))



import logging
import os
import subprocess
from pathlib import Path

import pandas as pd

from immuneML.data_model.SequenceParams import RegionType
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.environment.SequenceType import SequenceType
from immuneML.util.CompAIRRParams import CompAIRRParams


class CompAIRRHelper:

    @staticmethod
    def determine_compairr_path(compairr_path):
        if compairr_path is None:
            try:
                compairr_path = CompAIRRHelper.check_compairr_path("compairr")
            except Exception as e:
                compairr_path = CompAIRRHelper.check_compairr_path("/usr/local/bin/compairr")
        else:
            compairr_path = CompAIRRHelper.check_compairr_path(compairr_path)

        return compairr_path

    @staticmethod
    def check_compairr_path(compairr_path):
        required_major = 1
        required_minor = 3
        required_patch = 2

        try:
            compairr_result = subprocess.run([str(Path(compairr_path)), "--version"], capture_output=True)
            assert compairr_result.returncode == 0, "exit code was non-zero."
            output = str(compairr_result.stderr).split()
            major, minor, patch = output[1].split(".")

            mssg = f"CompAIRR version {required_major}.{required_minor}.{required_patch} or higher is required, found version {output[1]}"
            assert int(major) >= required_major, mssg
            if major == 1:
                assert int(minor) >= required_minor, mssg
                if minor == 3:
                    assert int(patch) >= {required_patch}, mssg
        except Exception as e:
            raise Exception(f"CompAIRRHelper: failed to call CompAIRR: {e}\n"
                            f"Please ensure the correct version of CompAIRR has been installed (version {required_major}.{required_minor}.{required_patch} or later), "
                            f"or provide the path to the CompAIRR executable.").with_traceback(e.__traceback__)

        return compairr_path

    @staticmethod
    def get_cmd_args(compairr_params: CompAIRRParams, input_file_list, result_path):
        indels_args = ["-i"] if compairr_params.indels else []
        frequency_args = ["-f"] if compairr_params.ignore_counts else []
        ignore_genes = ["-g"] if compairr_params.ignore_genes else []
        output_args = ["-o", str(result_path / compairr_params.output_filename), "-l", str(result_path / compairr_params.log_filename)]
        output_pairs = ['-p', str(result_path / compairr_params.pairs_filename)] if compairr_params.output_pairs else []
        cdr3_indicator = ['--cdr3'] if compairr_params.is_cdr3 else []
        command = '-m' if compairr_params.do_repertoire_overlap and not compairr_params.do_sequence_matching else '-x'

        return [str(compairr_params.compairr_path), command, "-d", str(compairr_params.differences), "-t", str(compairr_params.threads)] + \
               indels_args + frequency_args + ignore_genes + output_args + [str(file) for file in input_file_list] + output_pairs + cdr3_indicator

    @staticmethod
    def write_repertoire_file(repertoire_dataset=None, filename=None, compairr_params: CompAIRRParams = None,
                              repertoires: list = None, export_sequence_id: bool = False):
        mode = "w"
        header = True

        columns_in_order = []

        if repertoire_dataset is not None and repertoires is None:
            repertoires = repertoire_dataset.get_data()

        encoder_params = EncoderParams(region_type=RegionType.IMGT_CDR3 if compairr_params.is_cdr3 else RegionType.IMGT_JUNCTION)

        for ind, repertoire in enumerate(repertoires):
            repertoire_contents = CompAIRRHelper.get_repertoire_contents(repertoire, compairr_params, encoder_params,
                                                                         export_sequence_id)

            if ind == 0:
                columns_in_order = sorted(repertoire_contents.columns)

            repertoire_contents[columns_in_order].to_csv(filename, mode=mode, header=header, index=False, sep="\t")

            mode = "a"
            header = False

    @staticmethod
    def write_sequences_file(sequence_dataset, filename, compairr_params, repertoire_id="sequence_dataset"):
        compairr_data = {"junction_aa": [],
                         "repertoire_id": [],
                         "sequence_id": []}

        if not compairr_params.ignore_genes:
            compairr_data["v_call"] = []
            compairr_data["j_call"] = []

        if not compairr_params.ignore_counts:
            compairr_data["duplicate_count"] = []

        for sequence in sequence_dataset.get_data():
            compairr_data["junction_aa"].append(sequence.get_sequence(sequence_type=SequenceType.AMINO_ACID))

            assert sequence.sequence_id is not None, \
                (f"{CompAIRRHelper.__name__}: sequence identifiers must be set when exporting a sequence "
                 f"dataset for CompAIRR.")
            compairr_data["sequence_id"].append(sequence.sequence_id)
            compairr_data["repertoire_id"].append(repertoire_id)

            if not compairr_params.ignore_genes:
                compairr_data["v_call"].append(sequence.v_call)
                compairr_data["j_call"].append(sequence.j_call)

            if not compairr_params.ignore_counts:
                compairr_data["duplicate_count"].append(sequence.duplicate_count)

        df = pd.DataFrame(compairr_data)

        df.to_csv(filename, mode="w", header=True, index=False, sep="\t")

    @staticmethod
    def get_repertoire_contents(repertoire, compairr_params, encoder_params: EncoderParams, export_sequence_id=False):
        attributes = [encoder_params.get_sequence_field_name(), "duplicate_count"]
        attributes += [] if compairr_params.ignore_genes else ["v_call", "j_call"]
        repertoire_contents = repertoire.data.topandas()[attributes]
        repertoire_contents = pd.DataFrame({**repertoire_contents, "identifier": repertoire.identifier})
        if export_sequence_id:
            repertoire_contents['sequence_id'] = repertoire.data.sequence_id.tolist()

        check_na_rows = [encoder_params.get_sequence_field_name()]
        check_na_rows += [] if compairr_params.ignore_counts else ["duplicate_count"]
        check_na_rows += [] if compairr_params.ignore_genes else ["v_call", "j_call"]

        n_rows_before = len(repertoire_contents)

        repertoire_contents.dropna(inplace=True, subset=check_na_rows)

        if n_rows_before > len(repertoire_contents):
            logging.warning(
                f"CompAIRRHelper: removed {n_rows_before - len(repertoire_contents)} entries from repertoire {repertoire.identifier} due to missing values.")

        if compairr_params.ignore_counts:
            repertoire_contents["duplicate_count"] = 1
        else:
            repertoire_contents['duplicate_count'] = [count if count >= 0 else pd.NA for count in repertoire_contents['duplicate_count']]

        repertoire_contents.rename(columns={"identifier": "repertoire_id"},
                                   inplace=True)

        return repertoire_contents

    @staticmethod
    def verify_compairr_output_path(subprocess_result, compairr_params, result_path):
        output_file = result_path / compairr_params.output_filename

        if not output_file.is_file():
            raise RuntimeError(
                f"CompAIRRHelper: failed to calculate the distance matrix with CompAIRR ({compairr_params.compairr_path}). "
                f"The following error occurred: {subprocess_result.stderr}")

        if os.path.getsize(output_file) == 0:
            raise RuntimeError(
                f"CompAIRRHelper: failed to calculate the distance matrix with CompAIRR ({compairr_params.compairr_path}), output matrix is empty. "
                f"For details see the log file at {result_path / compairr_params.log_filename}")

        return output_file

    @staticmethod
    def read_compairr_output_file(output_file):
        return pd.read_csv(output_file, sep="\t", index_col=0)

    @staticmethod
    def process_compairr_output_file(subprocess_result, compairr_params, result_path):
        output_file = CompAIRRHelper.verify_compairr_output_path(subprocess_result, compairr_params, result_path)
        return CompAIRRHelper.read_compairr_output_file(output_file)



import logging
from pathlib import Path
from typing import Callable, Tuple

import pandas as pd
from sklearn import metrics
from sklearn.metrics import jaccard_score

from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.clustering_reports.ClusteringReport import ClusteringReport
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.instructions.clustering.ClusteringRunner import get_features
from immuneML.workflows.instructions.clustering.ClusteringState import ClusteringState
from immuneML.workflows.instructions.clustering.ValidationHandler import get_complementary_classifier
from immuneML.workflows.instructions.clustering.clustering_run_model import ClusteringSetting


class ClusteringStabilityReport(ClusteringReport):
    """
    Report that analyzes clustering stability by comparing results between discovery and validation datasets.
    The comparison uses a classifier-based approach where:
    1. A classifier is trained on discovery data using cluster assignments as labels
    2. Cluster assignments are predicted for validation data
    3. Predictions are compared with actual validation clustering results using the specified similarity metric

    This report can be used with the Clustering instruction under 'reports'.

    **Specification arguments:**

    - metric (str): Name of any clustering evaluation metric from sklearn.metrics that compares two sets of labels
      (e.g., adjusted_rand_score, jaccard_score, adjusted_mutual_info_score, normalized_mutual_info_score).
      If an invalid metric name is provided, defaults to adjusted_rand_score.


    YAML specification:

    .. indent with spaces
    .. code-block:: yaml

        my_clustering_instruction:
            type: Clustering
            reports:
                my_stability_report:
                    ClusteringStabilityReport:
                        metric: jaccard_score

    """

    DEFAULT_METRIC = "adjusted_rand_score"

    @classmethod
    def build_object(cls, **kwargs):
        location = "ClusteringStabilityReport"
        name = kwargs["name"] if "name" in kwargs else None
        similarity_metric = kwargs.get("similarity_metric", "ari")

        ParameterValidator.assert_type_and_value(similarity_metric, str, location, "similarity_metric")
        ParameterValidator.assert_in_valid_list(similarity_metric.lower(), ["ari", "jaccard"], location,
                                                "similarity_metric")

        return ClusteringStabilityReport(name=name, similarity_metric=similarity_metric)

    def __init__(self, similarity_metric: str, name: str = None, state: ClusteringState = None,
                 result_path: Path = None, number_of_processes: int = 1):
        super().__init__(name=name, result_path=result_path, number_of_processes=number_of_processes)
        self.state = state
        self.metric_fn, self.metric = self._get_similarity_metric(similarity_metric)
        self.result_name = "clustering_stability_analysis"
        self.desc = "Clustering Stability Report"

    def _get_similarity_metric(self, metric_name: str) -> Tuple[Callable, str]:
        try:
            metric_func = getattr(metrics, metric_name)
            if callable(metric_func):
                if metric_name == 'jaccard_score':
                    def jaccard_weighted(x, y):
                        return jaccard_score(x, y, average='weighted')

                    metric_func = jaccard_weighted
                    logging.info(f"Using Jaccard metric for clustering stability analysis with weighted average")
                return metric_func, metric_name
        except AttributeError:
            logging.warning(f"Invalid similarity metric '{metric_name}', defaulting to {self.DEFAULT_METRIC}")
            return getattr(metrics, self.DEFAULT_METRIC), self.DEFAULT_METRIC

    def _generate(self) -> ReportResult:
        """Generate the stability analysis report"""
        self.result_path = PathBuilder.build(self.result_path / self.name)
        stability_results = []

        # For each data split
        for split_id in range(self.state.config.split_config.split_count):
            # For each clustering setting
            for setting in self.state.config.clustering_settings:
                setting_key = setting.get_key()

                # Get discovery and validation results
                discovery_result = self._get_cluster_result('discovery', split_id, setting_key)
                validation_result = self._get_cluster_result('method_based_validation', split_id, setting_key)

                if discovery_result and validation_result:
                    discovery_clusters = discovery_result.item.predictions
                    validation_clusters = validation_result.item.predictions

                    # Get encoded data
                    discovery_data = get_features(discovery_result.item.dataset, setting)
                    validation_data = get_features(validation_result.item.dataset, setting)

                    if discovery_data is not None and validation_data is not None:
                        try:
                            # Train classifier on discovery data
                            clf = get_complementary_classifier(setting)
                            clf.fit(discovery_data, discovery_clusters)
                            predicted_clusters = clf.predict(validation_data)

                            # Calculate stability score
                            stability_score = self.metric_fn(validation_clusters, predicted_clusters)

                            # Add additional information
                            result_entry = {
                                'split': split_id + 1,
                                'clustering_setting': setting_key,
                                self.metric: stability_score,
                                'n_clusters_discovery': len(set(discovery_clusters)),
                                'n_clusters_validation': len(set(validation_clusters))
                            }

                            stability_results.append(result_entry)
                        except Exception as e:
                            logging.warning(f"Error calculating stability for split {split_id + 1}, setting {setting_key}: {str(e)}")

        if not stability_results:
            logging.warning("No stability results could be calculated. Check if validation data is available.")
            return ReportResult(
                name=f"{self.desc} ({self.name})",
                info="No stability results could be calculated. Check if validation data is available."
            )

        # Create report outputs
        df = pd.DataFrame(stability_results)

        # Save detailed results table
        table_path = self.result_path / f"{self.result_name}.csv"
        df.to_csv(table_path, index=False)
        table_output = ReportOutput(path=table_path, name="Stability scores per setting")

        return ReportResult(
            name=f"{self.desc} ({self.name})",
            info=f"Analysis of clustering stability between discovery and validation datasets using {self.metric}.",
            output_tables=[table_output]
        )

    def _get_cluster_result(self, analysis_type: str, split_id: int, setting_key: str):
        """Get clustering result for a specific analysis type, split, and setting"""
        try:
            cl_result = self.state.clustering_items[split_id]
            if hasattr(cl_result, analysis_type):
                analysis_result = getattr(cl_result, analysis_type)
                if analysis_result and setting_key in analysis_result.items:
                    return analysis_result.items[setting_key]
        except (IndexError, AttributeError, KeyError) as e:
            logging.warning(f"Could not retrieve clustering result for {analysis_type}, split {split_id}, setting {setting_key}: {str(e)}")
        return None

    def check_prerequisites(self):
        run_report = True

        if self.state is None or not isinstance(self.state, ClusteringState):
            logging.warning(f"{self.__class__.__name__} requires a valid state object. Report will not be created.")
            run_report = False

        return run_report



import abc

from immuneML.data_model.SequenceSet import ReceptorSequence
from immuneML.encodings.EncoderParams import EncoderParams


class SequenceEncodingStrategy(metaclass=abc.ABCMeta):

    @staticmethod
    @abc.abstractmethod
    def encode_sequence(sequence: ReceptorSequence, params: EncoderParams):
        pass

    @staticmethod
    @abc.abstractmethod
    def get_feature_names(params: EncoderParams):
        pass



import numpy as np
import pandas as pd

from immuneML import Constants
from immuneML.IO.dataset_import.DatasetImportParams import DatasetImportParams
from immuneML.data_model.SequenceParams import RegionType, Chain
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.util.ImportHelper import ImportHelper


class AdaptiveImportHelper:

    @staticmethod
    def preprocess_dataframe(dataframe: pd.DataFrame, params: DatasetImportParams):

        if "sequence" in dataframe.columns:
            dataframe['junction'] = [y[(84 - 3 * len(x)): 78] if is_valid_sequence_str(x) else ''
                                     for x, y in zip(dataframe['junction_aa'], dataframe['sequence'])]
            dataframe['cdr3'] = dataframe['junction'].str[3:-3]
        dataframe['cdr3_aa'] = dataframe['junction_aa'].str[1:-1]

        if "frame_type" in dataframe.columns:
            dataframe['vj_in_frame'] = (dataframe.frame_type.str.upper() == 'IN').astype(str).str[:1]
            dataframe['stop_codon'] = (dataframe.frame_type.str.upper() == 'STOP').astype(str).str[:1]
            dataframe['productive'] = dataframe.junction_aa.notnull().astype(str).str[:1]
            dataframe.drop(columns=['frame_type'], inplace=True)

        if 'duplicate_count' in dataframe.columns:
            dataframe.loc[dataframe['duplicate_count'].isna(), 'duplicate_count'] = -1
            dataframe.duplicate_count = dataframe.duplicate_count.astype(int)

        dataframe = AdaptiveImportHelper.parse_adaptive_germline_to_imgt(dataframe, params.organism)
        dataframe = set_locus_column(dataframe)

        return dataframe

    @staticmethod
    def parse_adaptive_germline_to_imgt(dataframe, organism):
        gene_name_replacement = pd.read_csv(
            EnvironmentSettings.root_path / "immuneML/IO/dataset_import/conversion/imgt_adaptive_conversion.csv")
        gene_name_replacement = gene_name_replacement[gene_name_replacement.Species == organism]
        gene_name_replacement = dict(zip(gene_name_replacement.Adaptive, gene_name_replacement.IMGT))

        # remove C and extra 0 from gene name but not from allele (e.g., TCRBV03-01*01 -> TRBV3-1*01) to follow IMGT
        # naming
        germline_value_replacement = {**{"TCRB": "TRB", "TCRA": "TRA"},
                                      **{f"-0{i}": f"-{str(i)}" for i in range(10)},
                                      **{f"J0": "J", "V0": "V"}}

        return AdaptiveImportHelper.parse_germline(dataframe, gene_name_replacement, germline_value_replacement)

    @staticmethod
    def parse_germline(dataframe: pd.DataFrame, gene_name_replacement: dict, germline_value_replacement: dict):

        for gene in ["v", "j"]:
            if f"{gene}_call" in dataframe.columns:

                dataframe = replace_nans_with_empty_str(dataframe, f"{gene}_call")
                dataframe[f"{gene}_call"].replace(gene_name_replacement, regex=True, inplace=True)
                dataframe[f"{gene}_call"].replace(germline_value_replacement, regex=True, inplace=True)

                if f"{gene}_allele" in dataframe.columns:
                    rows_to_add_allele = ~dataframe[f'{gene}_call'].str.contains("\\*") & dataframe[
                        f"{gene}_allele"].astype(
                        str).str.contains("[0-9]{2}")
                    dataframe.loc[rows_to_add_allele, f"{gene}_call"] = \
                        dataframe.loc[rows_to_add_allele, lambda df: [f"{gene}_call", f"{gene}_allele"]].agg('*'.join,
                                                                                                             axis=1)
            elif f"{gene}_gene" in dataframe.columns and f"{gene}_allele" in dataframe.columns:
                dataframe = parse_allele(dataframe, gene)
                dataframe = parse_gene_column(dataframe, gene, gene_name_replacement, germline_value_replacement)
                make_gene_call_from_gene_and_allele(dataframe, gene)
            elif f"{gene}_gene" in dataframe.columns:
                dataframe = parse_gene_column(dataframe, gene, gene_name_replacement, germline_value_replacement)
                dataframe.rename(columns={f'{gene}_gene': f'{gene}_call'}, inplace=True)

        dataframe.drop(columns=['v_gene', 'j_gene', 'v_allele', 'j_allele'], inplace=True, errors='ignore')
        return dataframe


def set_locus_column(df: pd.DataFrame):
    if 'locus' in df.columns:
        df.locus = [Chain.get_chain_value(item) for item in df.locus]
    elif 'v_call' in df.columns:
        df['locus'] = [Chain.get_chain_value(item[:3]) for item in df.v_call]
    elif 'j_call' in df.columns:
        df['locus'] = [Chain.get_chain_value(item[:3]) for item in df.j_call]
    return df


def make_gene_call_from_gene_and_allele(df: pd.DataFrame, gene: str):
    gene_ids = df[f"{gene}_gene"] != ''
    non_allele_ids = df[f"{gene}_allele"] == ''
    df[f"{gene}_call"] = df[[f"{gene}_gene", f"{gene}_allele"]].agg('*0'.join, axis=1)
    df.loc[~gene_ids, f'{gene}_call'] = ''
    gene_set_allele_not = np.logical_and(gene_ids, non_allele_ids)
    df.loc[gene_set_allele_not, f"{gene}_call"] = df.loc[gene_set_allele_not, f"{gene}_gene"]
    return df


def parse_allele(df: pd.DataFrame, gene: str):
    if f"{gene}_allele" in df.columns:
        df[f"{gene}_allele"] = df[f"{gene}_allele"].astype(str)
        allele_set = np.logical_and(df[f"{gene}_allele"] != 'nan', df[f"{gene}_allele"] != '-1.0')
        df.loc[allele_set, f"{gene}_allele"] = df.loc[allele_set, f"{gene}_allele"].str[:-2]
        df.loc[~allele_set, f"{gene}_allele"] = ''
    return df


def parse_gene_column(df: pd.DataFrame, gene, gene_name_replacement, germline_value_replacement):
    df[f"{gene}_gene"].replace(germline_value_replacement, regex=True, inplace=True)
    df[f"{gene}_gene"].replace(gene_name_replacement, regex=True, inplace=True)
    return df


def replace_nans_with_empty_str(df: pd.DataFrame, col: str):
    df[col].replace('nan', '', inplace=True)
    return df


def is_valid_sequence_str(x):
    return isinstance(x, str) and x not in ["unresolved", "no data", "na", "unknown", 'nan']



from typing import List

from bionumpy import AminoAcidEncoding, DNAEncoding
from bionumpy.bnpdataclass import bnpdataclass

from immuneML.data_model.SequenceSet import ReceptorSequence
from immuneML.environment.SequenceType import SequenceType


@bnpdataclass
class BackgroundSequences:
    sequence_aa: AminoAcidEncoding
    sequence: DNAEncoding
    v_call: str
    j_call: str
    region_type: str
    frame_type: str
    p_gen: float
    from_default_model: int
    duplicate_count: int
    locus: str

    def get_sequence(self, sequence_type: SequenceType = SequenceType.AMINO_ACID):
        if sequence_type == SequenceType.AMINO_ACID:
            return self.sequence_aa
        else:
            return self.sequence



import logging
import os
import shutil
from enum import Enum
from pathlib import Path
import numpy as np

from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.util.PathBuilder import PathBuilder
from immuneML.util.StringHelper import StringHelper


class Util:

    @staticmethod
    def to_dict_recursive(obj, base_path: Path):
        if not hasattr(obj, "__dict__"):
            if (isinstance(obj, Path) or isinstance(obj, str)) and os.path.isfile(obj):
                obj_abs_path = os.path.abspath(obj)
                base_abs_path = os.path.abspath(str(base_path))
                res_path = os.path.relpath(obj_abs_path, base_abs_path)
                return res_path
            elif hasattr(obj, "name"):
                return obj.name
            elif isinstance(obj, list):
                return [Util.to_dict_recursive(element, base_path) for element in obj]
            else:
                return obj if obj is not None else ""
        elif isinstance(obj, Enum):
            return str(obj)
        else:
            vars_obj = vars(obj)
            result = {
                attribute_name: Util.to_dict_recursive(vars_obj[attribute_name], base_path) for attribute_name in vars_obj.keys()
            }
            if isinstance(obj, ReportOutput):
                path = getattr(obj, "path")

                if path is not None:
                    result['is_download_link'] = True
                    result['download_link'] = os.path.relpath(path=getattr(obj, "path"), start=base_path)
                    result['file_name'] = getattr(obj, "name")

                    if any([ext == path.suffix for ext in ['.svg', '.jpg', '.png']]):
                        result['is_embed'] = False
                    else:
                        result['is_embed'] = True
                else:
                    result['is_download_link'] = False

            return result

    @staticmethod
    def get_css_content(css_path: Path):
        with css_path.open("rb") as file:
            content = file.read()
        return content

    @staticmethod
    def make_downloadable_zip(base_path: Path, path_to_zip: Path, filename: str = "") -> str:
        if filename == "":
            filename = "_".join(Path(os.path.relpath(str(path_to_zip), str(base_path)).replace(".", "")).parts)

        PathBuilder.build(base_path / "zip")
        zip_file_path = shutil.make_archive(base_name=base_path / f"zip/{filename}", format="zip", root_dir=str(path_to_zip))
        return zip_file_path

    @staticmethod
    def get_full_specs_path(base_path) -> str:
        specs_path = list(base_path.glob("../**/full*.yaml"))

        if len(specs_path) == 1:
            path_str = os.path.relpath(specs_path[0], base_path)
        else:
            path_str = ""

        return path_str

    @staticmethod
    def get_logfile_path(base_path) -> str:
        path_str = ""

        logfile_path = list(base_path.glob("../*log.txt"))

        if len(logfile_path) == 1:
            path_str = os.path.relpath(logfile_path[0], base_path)
        else:
            try:
                logger = logging.getLogger()
                if len(logger.handlers) == 1:
                    handler = logger.handlers[0]
                    if isinstance(handler, logging.handlers.FileHandler):
                        path_str = os.path.relpath(str(handler.baseFilename), base_path)
            except Exception:
                pass

        return path_str

    @staticmethod
    def get_table_string_from_csv_string(csv_string: str, separator: str = ",", has_header: bool = True) -> str:
        table_string = "<table>\n"
        for index, line in enumerate(csv_string.splitlines()):
            if index == 0 and has_header:
                table_string += "<thead>\n"
            table_string += "<tr>\n"
            for col in line.split(separator):
                table_string += f"<td>{col}</td>\n"
            table_string += "</tr>\n"
            if index == 0 and has_header:
                table_string += "</thead>\n"
        table_string += "</table>\n"
        return table_string

    @staticmethod
    def get_table_string_from_csv(csv_path: Path, separator: str = ",", has_header: bool = True) -> str:
        with csv_path.open("r") as file:
            table_string = Util.get_table_string_from_csv_string(file.read())
        return table_string

    @staticmethod
    def update_report_paths(report_result: ReportResult, path: Path) -> ReportResult:
        for attribute in vars(report_result):
            attribute_value = getattr(report_result, attribute)
            if isinstance(attribute_value, list):
                for output in attribute_value:
                    if isinstance(output, ReportOutput):
                        new_filename = "_".join([part for part in Path(os.path.relpath(path=str(output.path), start=str(path))).parts if part != ".."])
                        new_path = path / new_filename

                        if output.path != new_path:
                            shutil.copyfile(src=str(output.path), dst=new_path)
                            output.path = new_path
                    else:
                        logging.warning(f"HTML util: one of the report outputs was not returned properly from the "
                                        f"report {report_result.name}, and it will not be moved to HTML output folder.")

        return report_result

    @staticmethod
    def make_dataset_html_map(dataset, dataset_key="dataset"):
        return {f"{dataset_key}_name": dataset.name,
                f"{dataset_key}_type": StringHelper.camel_case_to_word_string(type(dataset).__name__),
                f"{dataset_key}_size": f"{dataset.get_example_count()} {type(dataset).__name__.replace('Dataset', 's').lower()}",
                f"{dataset_key}_labels": [{f"{dataset_key}_label_name": label_name,
                                           f"{dataset_key}_label_classes": ", ".join(str(class_name) for class_name in dataset.labels[label_name])}
                                                for label_name in dataset.get_label_names()],
                f"show_{dataset_key}_labels": len(dataset.get_label_names()) > 0}


import logging

import pandas as pd
import numpy as np


class EncodedData:
    """
    When a dataset is encoded, it is stored in an object of EncodedData class.

    Arguments:

      examples: a design matrix containing the encoded data. This is typically a numpy array, although other matrix formats such as scipy sparse matrix, pandas dataframe
                or pytorch tensors are also permitted as long as the numpy matrix can be retrieved using 'get_examples_as_np_matrix()'.
                The matrix is usually two-dimensional. The first dimension should be the examples, and the second (and higher) dimensions represent features.


      feature_names: a list of feature names. The length (dimensions) of this list should match the number of features in the examples matrix.

      feature_annotations: a data frame consisting of additional annotations for each feature. This can be used to add more information fields if feature_names is not sufficient. This data field is not used for machine learning, but may be used by some Reports.

      example_ids: a list of example (repertoire/sequence/receptor) IDs; it must be the same length as the example_count in the examples matrix. These can be retrieved using Dataset.get_example_ids()

      labels: a dict of labels where label names are keys and the values are lists of values for the label across examples: {'disease1': ['sick', 'healthy', 'sick']}
              During encoding, the labels can be computed using EncoderHelper.encode_dataset_labels()
      """

    def __init__(self, examples, labels: dict = None, example_ids: list = None, feature_names: list = None,
                 feature_annotations: pd.DataFrame = None, encoding: str = None, example_weights: list = None, info: dict = None,
                 dimensionality_reduced_data: np.ndarray = None):

        assert feature_names is None or examples.shape[1] == len(feature_names), \
            (f"EncodedData: the length of feature_names ({len(feature_names)}) must match the feature dimension of the "
             f"example matrix ({examples.shape[1]})")
        if feature_names is not None:
            assert feature_annotations is None or feature_annotations.shape[0] == len(feature_names) == examples.shape[1]
        if example_ids is not None and labels is not None:
            for label in labels.values():
                assert len(label) == len(example_ids), "EncodedData: there are {} labels, but {} examples"\
                    .format(len(label), len(example_ids))
                assert examples is None or len(example_ids) == examples.shape[0], (
                    "EncodedData: there are {} example ids, but {} examples."
                    .format(len(example_ids), examples.shape[0]))

            if example_weights is not None:
                assert len(example_weights) == len(example_ids)
        if examples is not None:
            assert all(len(labels[key]) == examples.shape[0] for key in labels.keys()) if labels is not None else True

            if example_weights is not None:
                assert len(example_weights) == examples.shape[0]

        self.examples = examples
        self.labels = labels
        self.example_ids = example_ids
        self.feature_names = feature_names
        self.feature_annotations = feature_annotations
        self.encoding = encoding
        self.example_weights = example_weights
        self.info = info
        self.dimensionality_reduced_data = dimensionality_reduced_data

    def get_examples_as_np_matrix(self):
        if isinstance(self.examples, np.ndarray):
            return self.examples
        elif isinstance(self.examples, pd.DataFrame):
            return self.examples.to_numpy()
        try:
            from scipy.sparse import issparse
            if issparse(self.examples):
                return self.examples.toarray()
        except ImportError as e:
            logging.warning(f"{EncodedData.__name__}: scipy could not be imported.")

        try:
            import torch
            if torch.is_tensor(self.examples):
                return self.examples.numpy()
        except ImportError as e:
            logging.warning(f"{EncodedData.__name__}: torch could not be imported.")

        raise ValueError(f"EncodedData: examples matrix of type '{type(self.examples)}' cannot be converted to a numpy matrix.")


# quality: gold
import os
from pathlib import Path

from immuneML.caching.CacheType import CacheType
from immuneML.environment.Constants import Constants
from immuneML.environment.SequenceType import SequenceType
from immuneML.util.Logger import print_log
from immuneML.util.PathBuilder import PathBuilder


class EnvironmentSettings:
    """
    Class containing environment variables, like receptor_sequence type,
    root path etc.
    """

    sequence_type = SequenceType.AMINO_ACID
    root_path = Path(os.path.normpath(os.path.dirname(os.path.abspath(__file__)) + "/../../") + "/")
    default_params_path = root_path / "immuneML/config/default_params"
    tmp_test_path = root_path / "test/tmp"
    default_analysis_path = root_path / "analysis_runs"
    cache_path = root_path / "cache"
    tmp_cache_path = tmp_test_path / "cache"
    html_templates_path = root_path / "immuneML/presentation/html/templates"
    specs_docs_path = root_path / "docs/specs"
    source_docs_path = root_path / "docs/source"
    max_sequence_length = 20
    low_memory = True
    compairr_paths = [Path("/usr/local/bin/compairr"), Path("./compairr/src/compairr"),
                      root_path / 'compairr/src/compairr', root_path / 'compairr']

    @staticmethod
    def reset_cache_path():
        EnvironmentSettings.cache_path = EnvironmentSettings.root_path / "cache"
        del os.environ[Constants.CACHE_PATH]

    @staticmethod
    def set_cache_path(path: Path):
        EnvironmentSettings.cache_path = Path(path)
        PathBuilder.build(path)
        os.environ[Constants.CACHE_PATH] = str(EnvironmentSettings.cache_path)
        print_log(f"Setting temporary cache path to {path}", include_datetime=True)

    @staticmethod
    def get_cache_type():
        if Constants.CACHE_TYPE not in os.environ:
            os.environ[Constants.CACHE_TYPE] = CacheType.PRODUCTION.name
        return CacheType[os.environ[Constants.CACHE_TYPE].upper()]

    @staticmethod
    def get_cache_path(cache_type: CacheType = None):
        cache_type = EnvironmentSettings.get_cache_type() if cache_type is None else cache_type
        if cache_type == CacheType.PRODUCTION:
            return EnvironmentSettings.cache_path if Constants.CACHE_PATH not in os.environ else Path(os.environ[Constants.CACHE_PATH])
        elif cache_type == CacheType.TEST:
            return EnvironmentSettings.tmp_cache_path
        else:
            raise RuntimeError("Cache is not set up.")

    @staticmethod
    def set_sequence_type(sequence_type: SequenceType):
        EnvironmentSettings.sequence_type = sequence_type

    @staticmethod
    def get_sequence_type() -> SequenceType:
        return EnvironmentSettings.sequence_type

    @staticmethod
    def get_sequence_alphabet(sequence_type: SequenceType = None):
        """
        :return: alphabetically sorted receptor_sequence alphabet
        """
        seq_type = sequence_type if sequence_type is not None else EnvironmentSettings.sequence_type
        if seq_type == SequenceType.AMINO_ACID:
            alphabet = list("ACDEFGHIKLMNPQRSTVWY")
            alphabet.sort()
        elif seq_type == SequenceType.NUCLEOTIDE:
            alphabet = list("ACGT")
            alphabet.sort()
        else:
            raise RuntimeError("EnvironmentSettings: the sequence alphabet cannot be obtained if sequence_type was "
                               "not set properly. Expected AMINO_ACID or NUCLEOTIDE, but got {seq_type} instead.")
        return alphabet



from pathlib import Path

from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.ml_methods.generative_models.GenerativeModel import GenerativeModel
from immuneML.reports.Report import Report


class TrainGenModelReport(Report):
    """
    TrainGenModel reports show some type of features or statistics comparing two datasets: the original and generated
    one, potentially in combination with the trained model. These reports can only be used inside TrainGenModel
    instruction with the aim of comparing two datasets: the dataset used to train a generative model and the dataset
    created from the trained model.
    """

    def __init__(self, original_dataset: Dataset = None, generated_dataset: Dataset = None, result_path: Path = None,
                 name: str = None, number_of_processes: int = 1, model: GenerativeModel = None):
        """
        The arguments defined below are set at runtime by the instruction.
        Concrete classes inheriting DataComparisonReport may include additional parameters that will be set by the user
        in the form of input arguments (e.g., from the YAML file).

        Args:

            original_dataset (Dataset): a dataset object (can be repertoire, receptor or sequence dataset, depending
            on the specific report) provided as input to the TrainGenModel instruction

            generated_dataset (Dataset): a dataset object as produced from the generative model after being trained on
            the original dataset

            result_path (Path): location where the results (plots, tables, etc.) will be stored

            name (str): user-defined name of the report used in the HTML overview automatically generated by the
            platform from the key used to define the report in the YAML

            number_of_processes (int): how many processes should be created at once to speed up the analysis.
            For personal machines, 4 or 8 is usually a good choice.

            model (GenerativeModel): trained generative model from the instruction

        """
        super().__init__(name=name, number_of_processes=number_of_processes)
        self.original_dataset = original_dataset
        self.generated_dataset = generated_dataset
        self.model = model
        self.result_path = result_path

    @staticmethod
    def get_title():
        return "TrainGenModel reports"



import copy
import itertools
import math
import os
import random
from dataclasses import fields, field
from itertools import chain
from multiprocessing import Pool
from pathlib import Path
from typing import List, Dict, Tuple, Union

import bionumpy as bnp
import dill
import numpy as np
import pandas as pd
from bionumpy import as_encoded_array
from bionumpy.bnpdataclass import BNPDataClass
from bionumpy.encoded_array import BaseEncoding

from immuneML.IO.dataset_export.AIRRExporter import AIRRExporter
from immuneML.app.LigoApp import SimError
from immuneML.data_model.SequenceSet import Repertoire
from immuneML.data_model.bnp_util import bnp_write_to_file, write_dataset_yaml
from immuneML.data_model.datasets.ElementDataset import ReceptorDataset
from immuneML.data_model.datasets.ElementDataset import SequenceDataset
from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.environment.SequenceType import SequenceType
from immuneML.ml_methods.generative_models.BackgroundSequences import BackgroundSequences
from immuneML.simulation.LigoSimState import LigoSimState
from immuneML.simulation.SimConfig import SimConfig
from immuneML.simulation.SimConfigItem import SimConfigItem
from immuneML.simulation.implants.Signal import Signal
from immuneML.simulation.simulation_strategy.ImplantingStrategy import ImplantingStrategy
from immuneML.simulation.util.bnp_util import merge_dataclass_objects
from immuneML.simulation.util.util import get_bnp_data, make_annotated_dataclass, \
    get_sequence_per_signal_count, \
    update_seqs_without_signal, update_seqs_with_signal, check_iteration_progress, make_sequence_paths, \
    needs_seqs_with_signal, \
    check_sequence_count, make_repertoire_from_sequences, get_no_signal_sequences, get_signal_sequences, \
    annotate_sequences, get_signal_sequence_count, filter_sequences_by_length, prepare_data_for_airr_seq_set
from immuneML.util.ExporterHelper import ExporterHelper
from immuneML.util.Logger import print_log
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.instructions.Instruction import Instruction
from immuneML.workflows.instructions.ligo_simulation.runtime_reports import make_p_gen_histogram_plot


class LigoSimInstruction(Instruction):
    """
    LIgO simulation instruction creates a synthetic dataset from scratch based on the generative model and a set of signals provided by
    the user.

    **Specification arguments:**

    - simulation (str): a name of a simulation object containing a list of SimConfigItem as specified under definitions key; defines how to combine signals with simulated data; specified under definitions

    - sequence_batch_size (int): how many sequences to generate at once using the generative model before checking for signals and filtering

    - max_iterations (int): how many iterations are allowed when creating sequences

    - export_p_gens (bool): whether to compute generation probabilities (if supported by the generative model) for sequences and include them as part of output

    - number_of_processes (int): determines how many simulation items can be simulated in parallel


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        instructions:
            my_simulation_instruction: # user-defined name of the instruction
                type: LIgOSim # which instruction to execute
                simulation: sim1
                sequence_batch_size: 1000
                max_iterations: 1000
                export_p_gens: False
                number_of_processes: 4

    """

    def __init__(self, simulation: SimConfig, signals: List[Signal], name: str,
                 sequence_batch_size: int, max_iterations: int, number_of_processes: int, export_p_gens: bool = None):

        self.state = LigoSimState(simulation=simulation, signals=signals, name=name)
        self._number_of_processes = number_of_processes
        self._sequence_batch_size = sequence_batch_size
        self._max_iterations = max_iterations
        self._export_p_gens = export_p_gens

        self._use_p_gens = self.state.simulation.keep_p_gen_dist and \
                           all(sim_item.generative_model.can_compute_p_gens() for sim_item in
                               self.state.simulation.sim_items)

        self._export_observed_signals = any(
            [it.false_negative_prob_in_receptors > 0 or it.false_positive_prob_in_receptors > 0
             for it in self.state.simulation.sim_items])

        self._noise_fields = [(f"observed_{s.id}", int) for s in
                              self.state.signals] if self._export_observed_signals else []

        if isinstance(self.state.simulation.simulation_strategy, ImplantingStrategy):
            implanting_fields = [('original_sequence', str, ''),
                                 ('original_p_gen', float, -1.)]
        else:
            implanting_fields = []

        self._annotation_fields = sorted(
            [(signal.id, int, 0) for signal in self.state.signals] + [('signals_aggregated', str, '')] +
            [(f"{signal.id}_positions", str, '') for signal in self.state.signals] + self._noise_fields,
            key=lambda x: x[0]) + implanting_fields

        self._custom_fields = dill.dumps(self._annotation_fields + [('p_gen', float, -1.), ('from_default_model', int, 1)])
        self._annotation_fields = dill.dumps(self._annotation_fields)

        self._background_fields = [(f.name, f.type) for f in fields(BackgroundSequences)]

    @property
    def sequence_type(self) -> SequenceType:
        return self.state.simulation.sequence_type

    @property
    def custom_fields(self):
        return dill.loads(self._custom_fields)

    @property
    def annotation_fields(self):
        return dill.loads(self._annotation_fields)

    @property
    def _annotated_dataclass(self):
        return make_annotated_dataclass(self.annotation_fields, self.state.signals)

    MIN_RANGE_PROBABILITY = 1e-5

    def run(self, result_path: Path):
        self.state.result_path = PathBuilder.build(result_path / self.state.name)

        self._simulate_dataset()
        self._export_dataset()

        return self.state

    def _export_dataset(self):

        exporter_output = ExporterHelper.export_dataset(self.state.resulting_dataset, [AIRRExporter],
                                                        self.state.result_path, omit_columns=['from_default_model'])

        self.state.formats = exporter_output['formats']
        self.state.paths = exporter_output['paths']

    def _simulate_dataset(self):

        labels = {**{signal.id: [True, False] for signal in self.state.signals},
                  **{'species': self.state.simulation.species}}

        if self.state.simulation.is_repertoire:
            self._simulate_repertoire_dataset(labels)
        else:
            self._simulate_receptor_dataset(labels)

    def _simulate_repertoire_dataset(self, labels: dict):

        repertoires = self._parallel_create_seq_sets()
        repertoires = list(itertools.chain.from_iterable(repertoires.values()))

        self.state.resulting_dataset = RepertoireDataset.build_from_objects(labels=labels, repertoires=repertoires,
                                                                            name='simulated_dataset',
                                                                            metadata_path=self.state.result_path / 'simulated_dataset_metadata.csv',
                                                                            path=self.state.result_path)

    def _simulate_receptor_dataset(self, labels: dict):

        receptor_files = self._parallel_create_seq_sets()

        data_filename = self.state.result_path / 'simulated_dataset.tsv'
        metadata_filename = self.state.result_path / 'simulated_dataset.yaml'

        for file in receptor_files.values():
            df = pd.read_csv(str(file), sep='\t', dtype={'junction': str, 'junction_aa': str, 'cdr3': str, 'cdr3_aa': str},
                             keep_default_na=False)
            df = prepare_data_for_airr_seq_set(df)
            df = df[sorted(df.columns)]
            if data_filename.exists():
                df.to_csv(str(data_filename), sep='\t', index=False, mode='a', header=False)
            else:
                df.to_csv(str(data_filename), sep='\t', index=False, mode='w', header=True)

        type_dict = {k: v for k, v, default_value in self.custom_fields}

        dataset_name = 'simulated_dataset'

        dataset_cls = ReceptorDataset if self.state.simulation.paired else SequenceDataset

        metadata_yaml = dataset_cls.create_metadata_dict(dataset_class=dataset_cls,
                                                         filename=data_filename,
                                                         type_dict=type_dict,
                                                         name=dataset_name,
                                                         labels=labels) # todo: identifier not explicitly passed on to dataset

        write_dataset_yaml(metadata_filename, metadata_yaml)

        self.state.resulting_dataset = dataset_cls.build(data_filename,
                                                         metadata_filename=metadata_filename,
                                                         name=dataset_name, labels=labels)

    def _parse_example_output(self, result) -> dict:
        if self.state.simulation.is_repertoire:
            examples = {k: [dill.loads(el) for el in v] for d in result for k, v in d.items()}
        else:
            examples = {k: v for d in result for k, v in d.items()}
        return examples

    def _parallel_create_seq_sets(self) -> dict:
        if self._number_of_processes > 1:
            chunk_size = math.ceil(len(self.state.simulation.sim_items) / self._number_of_processes)

            with Pool(processes=max(self._number_of_processes, len(self.state.simulation.sim_items))) as pool:
                result = pool.map(self._create_sequence_sets,
                                  [dill.dumps(item) for item in self.state.simulation.sim_items],
                                  chunksize=chunk_size)
                examples = self._parse_example_output(result)
        else:
            examples = {}
            for item in self.state.simulation.sim_items:
                res = self._create_sequence_sets(item)
                examples = {**examples, **{k: [dill.loads(el) for el in v] for k, v in res.items()}}

        return examples

    def _pair_examples(self, examples: Dict[str, list], path: Path) -> list:
        paired_examples = []
        pair_func = self._pair_repertoires if self.state.simulation.is_repertoire else self._pair_sequences
        for paired_item1, paired_item2 in self.state.simulation.paired:
            paired_examples.extend(pair_func(examples[paired_item1], examples[paired_item2], path))
        return paired_examples

    def _pair_repertoires(self, repertoires1: list, repertoires2: list, path: Path) -> List[Repertoire]:
        assert len(repertoires1) == len(repertoires2), \
            (f"{LigoSimInstruction.__name__}: cannot create paired repertoires, number of repertoires per "
             f"chain don't match: {len(repertoires1)} and {len(repertoires2)}.")
        PathBuilder.build(path)
        paired_repertoires = []
        for i in range(len(repertoires1)):
            repertoire = self._pair_two_repertories(repertoires1[i], repertoires2[i], path)
            paired_repertoires.append(repertoire)
        return paired_repertoires

    def _pair_two_repertories(self, repertoire1: Repertoire, repertoire2: Repertoire, path: Path) -> Repertoire:
        assert repertoire1.get_element_count() == repertoire2.get_element_count(), \
            (f"{LigoSimInstruction.__name__}: cannot pair repertoires {repertoire1.identifier} and "
             f"{repertoire2.identifier}, they have different number of sequences: {repertoire1.get_element_count()} "
             f"and {repertoire2.get_element_count()}.")

        data = []
        for sequences in [repertoire1.data, repertoire2.data]:
            sequences.sequence_id = bnp.as_encoded_array(
                [f"{sid}_{locus}" for sid, locus in zip(sequences.sequence_id.tolist(), sequences.locus.tolist())])
            sequences.cell_id = bnp.as_encoded_array([f"cell_{index}" for index in range(len(sequences))])
            data.append(sequences)

        return Repertoire.build_from_dc_object(path, {**repertoire1.metadata, **repertoire2.metadata}, None, None,
                                               merge_dataclass_objects(data))

    def _pair_sequences(self, sequences1, sequences2, path: Path = None) -> BNPDataClass:
        assert len(sequences1) == len(sequences2), (f"{LigoSimInstruction.__name__}: could not create paired dataset, "
                                                    f"the number of sequences in two simulation items did not match.")

        cell_ids = [f'cell_{i}' for i in range(len(sequences1))]
        sequences1.cell_id = as_encoded_array(cell_ids, BaseEncoding)
        random.shuffle(cell_ids)
        sequences2.cell_id = as_encoded_array(cell_ids, BaseEncoding)

        return merge_dataclass_objects([sequences1, sequences2])

    def _create_sequence_sets(self, item_in) -> Dict[str, Union[List[Repertoire], Path]]:

        item = dill.loads(item_in) if isinstance(item_in, bytes) else item_in

        if self.state.simulation.is_repertoire:
            res = self._create_repertoires(item)
        else:
            res = self._create_receptors(item)

        return {item.name: res}

    def _create_receptors(self, sim_item: SimConfigItem) -> Path:

        assert len(sim_item.signals) in [0,
                                         1], f"{LigoSimInstruction.__name__}: for sequence datasets, only 0 or 1 signal or a signal pair per " \
                                             f"sequence are supported, but {len(sim_item.signals)} were specified."

        sequence_paths = self._gen_necessary_sequences(self.state.result_path, sim_item)
        sequences = None

        if len(sim_item.signals) == 0:
            sequences = get_bnp_data(sequence_paths['no_signal'], self._annotated_dataclass)
        else:
            for signal in sim_item.signals:
                signal_sequences = get_bnp_data(sequence_paths[signal.id], self._annotated_dataclass)
                sequences = signal_sequences if sequences is None else merge_dataclass_objects(
                    [sequences, signal_sequences])

        sequences = self._compute_p_gens_for_export(sequences, sim_item)

        bnp_write_to_file(self.state.result_path / f"tmp_{sim_item.name}.tsv", sequences)

        return self.state.result_path / f"tmp_{sim_item.name}.tsv"

    def _compute_p_gens_for_export(self, sequences, sim_item: SimConfigItem):
        if self._export_p_gens:
            sequences = self._update_sequences_with_missing_p_gens(sequences, sim_item)
            if isinstance(self.state.simulation.simulation_strategy, ImplantingStrategy):
                sequences = self._update_sequences_with_missing_p_gens(sequences, sim_item,
                                                                       'original_sequence', 'original_p_gen')

        return sequences

    def _create_repertoires(self, item: SimConfigItem) -> list:
        path = PathBuilder.build(self.state.result_path / item.name)
        sequence_paths = self._gen_necessary_sequences(path, sim_item=item)

        repertoires = []
        used_seq_count = {**{'no_signal': 0}, **{signal.id: 0 for signal in item.signals}}
        repertoires_path = PathBuilder.build(path / "repertoires")

        for i in range(item.number_of_examples):
            seqs_no_signal_count = item.receptors_in_repertoire_count - sum(
                get_signal_sequence_count(1, proportion, item.receptors_in_repertoire_count)
                for _, proportion in item.signal_proportions.items())

            sequences, used_seq_count = get_signal_sequences(self._annotated_dataclass, used_seq_count, item,
                                                             sequence_paths)

            sequences, used_seq_count = get_no_signal_sequences(sequences=sequences, used_seq_count=used_seq_count,
                                                                seqs_no_signal_count=seqs_no_signal_count,
                                                                bnp_data_class=self._annotated_dataclass,
                                                                sequence_paths=sequence_paths,
                                                                sim_item=item)

            check_sequence_count(item, sequences)

            sequences = self._compute_p_gens_for_export(sequences, item)

            repertoire = make_repertoire_from_sequences(sequences, repertoires_path, item, self.state.signals)
            repertoires.append(dill.dumps(repertoire))

        return repertoires

    def _gen_necessary_sequences(self, base_path: Path, sim_item: SimConfigItem) -> Dict[str, Path]:

        if sim_item.seed is not None:
            np.random.seed(sim_item.seed)

        path = PathBuilder.build(base_path / sim_item.name)
        seqs_per_signal_count = get_sequence_per_signal_count(sim_item)
        seq_paths = make_sequence_paths(path, sim_item.signals)
        iteration = 1

        while sum(seqs_per_signal_count.values()) > 0 and iteration < self._max_iterations:
            sequences = self._make_background_sequences(path, iteration, sim_item, seqs_per_signal_count,
                                                        need_background_seqs=iteration == 1 and self.state.simulation.keep_p_gen_dist)

            if self.state.simulation.keep_p_gen_dist and sim_item.generative_model.can_compute_p_gens() and iteration == 1:
                self._make_p_gen_histogram(sequences, sim_item.name, path)
                print_log(
                    f"Computed a histogram from the first batch of background sequences for {sim_item.name}, available at: {str(path)}",
                    include_datetime=True)

            sequences = filter_sequences_by_length(sequences, sim_item, self.sequence_type)

            sequences = annotate_sequences(sequences, self.sequence_type == SequenceType.AMINO_ACID, self.state.signals,
                                           self._annotated_dataclass, sim_item.name, sim_item.generative_model.region_type)

            sequences = self.state.simulation.simulation_strategy.process_sequences(sequences, copy.deepcopy(
                seqs_per_signal_count), self._use_p_gens, self.sequence_type, sim_item, self.state.signals,
                                                                                    self.state.simulation.remove_seqs_with_signals,
                                                                                    implanting_scaling_factor=self.state.simulation.implanting_scaling_factor)

            if sequences is not None and len(sequences) > 0:

                if self.state.simulation.keep_p_gen_dist and sim_item.generative_model.can_compute_p_gens():
                    sequences = self._filter_using_p_gens(sequences, sim_item)

                seqs_per_signal_count['no_signal'] = update_seqs_without_signal(seqs_per_signal_count['no_signal'],
                                                                                sequences, seq_paths['no_signal'])
                seqs_per_signal_count = update_seqs_with_signal(copy.deepcopy(seqs_per_signal_count), sequences,
                                                                self.state.signals, sim_item.signals,
                                                                seq_paths)

            print_log(
                f"Finished iteration {iteration} in {sim_item.name}: remaining sequence count per signal for {sim_item.name}: "
                f"{seqs_per_signal_count}" if sum(
                    seqs_per_signal_count.values()) > 0 else f"{sim_item.name} simulation finished", True)
            check_iteration_progress(iteration, self._max_iterations)
            iteration += 1

        if iteration == self._max_iterations and sum(seqs_per_signal_count.values()) != 0:
            raise SimError(
                f"{LigoSimInstruction.__name__}: maximum iterations were reached, but the simulation could not finish "
                f"with parameters: {vars(self.state.simulation)}.\n")

        return seq_paths

    def _make_background_sequences(self, path, iteration: int, sim_item: SimConfigItem, sequence_per_signal_count: dict,
                                   need_background_seqs: bool) -> BackgroundSequences:
        sequence_path = PathBuilder.build(path / f"gen_model/") / f"tmp_{iteration}.tsv"

        v_genes = sorted(list(set(chain(signal.v_call for signal in sim_item.signals if signal.v_call is not None))))
        j_genes = sorted(list(set(chain(signal.j_call for signal in sim_item.signals if signal.j_call is not None))))

        if sequence_per_signal_count['no_signal'] > 0 or need_background_seqs or (
                len(v_genes) == 0 and len(j_genes) == 0) \
                or not sim_item.generative_model.can_generate_from_skewed_gene_models():
            sim_item.generative_model.generate_sequences(self._sequence_batch_size, seed=sim_item.seed,
                                                         path=sequence_path,
                                                         sequence_type=self.sequence_type,
                                                         compute_p_gen=self._use_p_gens)

            print_log(f"Generated {self._sequence_batch_size} background sequences, stored at {sequence_path}.", True)

        skew_model_for_signal = needs_seqs_with_signal(sequence_per_signal_count)

        if sim_item.generative_model.can_generate_from_skewed_gene_models() and skew_model_for_signal and (
                len(v_genes) > 0 or len(j_genes) > 0):
            sim_item.generative_model.generate_from_skewed_gene_models(v_genes=v_genes, j_genes=j_genes,
                                                                       seed=sim_item.seed, path=sequence_path,
                                                                       sequence_type=self.sequence_type,
                                                                       batch_size=self._sequence_batch_size,
                                                                       compute_p_gen=self._use_p_gens)

            print_log(
                f"Generated {self._sequence_batch_size} sequences from skewed model for given V/J genes at {sequence_path}.",
                True)

        data = get_bnp_data(sequence_path, BackgroundSequences)

        os.remove(sequence_path)
        print_log(f"Prepared sequences for processing and removed temporary file {sequence_path}.", True)

        return data

    def _make_p_gen_histogram(self, sequences: BackgroundSequences, sim_item_name: str, path: Path):

        log_p_gens = np.log10(sequences[sequences.from_default_model.astype(bool)].p_gen[
                                  np.nonzero(sequences[sequences.from_default_model.astype(bool)].p_gen)])

        hist, self.state.p_gen_bins[sim_item_name] = np.histogram(log_p_gens, density=False,
                                                                  bins=self.state.simulation.p_gen_bin_count)
        self.state.target_p_gen_histogram[sim_item_name] = hist / np.sum(sequences.from_default_model)

        zero_regions = self.state.target_p_gen_histogram[sim_item_name] == 0
        self.state.target_p_gen_histogram[sim_item_name][zero_regions] = ImplantingStrategy.MIN_RANGE_PROBABILITY
        self.state.target_p_gen_histogram[sim_item_name][np.logical_not(zero_regions)] -= \
            ImplantingStrategy.MIN_RANGE_PROBABILITY * (np.sum(zero_regions) + 1) / np.sum(np.logical_not(zero_regions))

        make_p_gen_histogram_plot(self.state.target_p_gen_histogram[sim_item_name],
                                  self.state.p_gen_bins[sim_item_name], path,
                                  str(LigoSimInstruction.MIN_RANGE_PROBABILITY))

    def _update_sequences_with_missing_p_gens(self, sequences: BackgroundSequences, sim_item: SimConfigItem,
                                              sequence_field: str = None, p_gen_field: str = "p_gen"):
        if np.any(getattr(sequences, p_gen_field) == -1):
            missing_p_gens = getattr(sequences, p_gen_field) == -1
            p_gens = getattr(sequences, p_gen_field)
            p_gens[missing_p_gens] = sim_item.generative_model.compute_p_gens(sequences[missing_p_gens],
                                                                              self.state.simulation.sequence_type,
                                                                              sequence_field)
            setattr(sequences, p_gen_field, p_gens)
        return sequences

    def _filter_using_p_gens(self, sequences: BackgroundSequences, sim_item: SimConfigItem) \
            -> Tuple[BNPDataClass, dict]:
        sequences = self._update_sequences_with_missing_p_gens(sequences, sim_item)
        with np.errstate(divide='ignore'):
            p_gens = np.log10(sequences.p_gen)

        p_gen_bins = self.state.p_gen_bins[sim_item.name]
        hist = np.concatenate(
            [[LigoSimInstruction.MIN_RANGE_PROBABILITY], self.state.target_p_gen_histogram[sim_item.name],
             [LigoSimInstruction.MIN_RANGE_PROBABILITY]])

        seq_keep_prob = hist[np.digitize(p_gens, p_gen_bins)]
        keep_sequences = np.random.uniform(0, 1, len(sequences)) <= seq_keep_prob

        print_log(
            f"Removed {len(sequences) - sum(keep_sequences)} out of {len(sequences)} sequences from the batch when filtering by p_gens.",
            True)

        return sequences[keep_sequences]



from pathlib import Path
from typing import List

from immuneML.hyperparameter_optimization.states.TrainMLModelState import TrainMLModelState
from immuneML.reports.Report import Report


class MultiDatasetReport(Report):
    '''
    Multi dataset reports are special reports that can be specified when running immuneML with the :py:obj:`~immuneML.api.aggregated_runs.MultiDatasetBenchmarkTool.MultiDatasetBenchmarkTool`.
    See Manuscript use case 1: :ref:`Robustness assessment` for an example.

    When running the :py:obj:`~immuneML.api.aggregated_runs.MultiDatasetBenchmarkTool.MultiDatasetBenchmarkTool`, multi dataset reports can be specified under 'benchmark_reports'.
    Example:

    .. indent with spaces
    .. code-block:: yaml

        my_instruction:
            type: TrainMLModel
            benchmark_reports:
                - my_benchmark_report
            # other parameters...
    '''
    DOCS_TITLE = "Multi dataset reports"

    def __init__(self, instruction_states: List[TrainMLModelState] = None, name: str = None, result_path: Path = None, number_of_processes: int = 1):
        '''
         When using the reports with MultiDatasetBenchmarkTool, the arguments defined below are set at runtime by the instruction.
        Concrete classes inheriting MultiDatasetReport may include additional parameters that will be set by the user in the form of input arguments.

        name (str): user-defined name of the report used in the HTML overview automatically generated by the platform
        result_path (Path): location where the report results will be stored
        instruction_states (list): a list of states for each instruction that was run as a part of the tool, e.g., TrainMLModelState objects
        number_of_processes (int): how many processes should be created at once to speed up the analysis. For personal machines, 4 or 8 is usually a good choice.
        '''
        super().__init__(name=name, result_path=result_path, number_of_processes=number_of_processes)
        self.instruction_states = instruction_states



from pathlib import Path
import pandas as pd

from immuneML.data_model.SequenceParams import RegionType
from immuneML.environment.SequenceType import SequenceType


class ParameterValidator:

    @staticmethod
    def assert_any_value_present(values: list, expected_values: list, location: str, parameter_name: str):
        assert any(exp_val in values for exp_val in
                   expected_values), f"{location}: expected at least one of values {expected_values} for " \
                                     f"parameter {parameter_name} to be set, but none were found in {values}."

    @staticmethod
    def assert_keys_present(values: list, expected_values: list, location: str, parameter_name: str):
        for value in expected_values:
            assert value in values, f"{location}: expected {value} to be set for {parameter_name}, but got {str(values)[1:-1]} instead."

    @staticmethod
    def assert_all_in_valid_list(values: list, valid_values: list, location: str, parameter_name: str):
        for value in values:
            ParameterValidator.assert_in_valid_list(value, valid_values, location, parameter_name)

    @staticmethod
    def assert_in_valid_list(value, valid_values: list, location: str, parameter_name: str):
        assert value in valid_values, \
            f"{location}: {value} is not a valid value for parameter {parameter_name}. Valid values are {str(valid_values)[1:-1]}."

    @staticmethod
    def assert_all_type_and_value(values, parameter_type, location: str, parameter_name: str, min_inclusive=None,
                                  max_inclusive=None):
        for value in values:
            ParameterValidator.assert_type_and_value(value, parameter_type, location, parameter_name, min_inclusive,
                                                     max_inclusive)

    @staticmethod
    def assert_type_and_value(value, parameter_type, location: str, parameter_name: str,
                              min_inclusive=None, max_inclusive=None,
                              min_exclusive=None, max_exclusive=None, exact_value=None):
        type_name = " or ".join([t.__name__ for t in parameter_type]) if type(
            parameter_type) is tuple else parameter_type.__name__

        base_mssg = f"{location}: {value} is not a valid value for parameter {parameter_name}. "

        assert isinstance(value,
                          parameter_type), f"{base_mssg}It has to be of type {type_name}, but is now of type {type(value).__name__}."

        if min_inclusive is not None:
            assert value >= min_inclusive, base_mssg + f"It has to be greater or equal to {min_inclusive}."

        if max_inclusive is not None:
            assert value <= max_inclusive, base_mssg + f"It has to be less or equal to {max_inclusive}."

        if min_exclusive is not None:
            assert value > min_exclusive, base_mssg + f"It has to be greater than {min_inclusive}."

        if max_exclusive is not None:
            assert value < max_exclusive, base_mssg + f"It has to be less than {max_inclusive}."

        if exact_value is not None:
            assert value == exact_value, base_mssg + f"It has to be equal to {exact_value}."

    @staticmethod
    def assert_keys(keys, valid_keys, location: str, parameter_name: str, exclusive: bool = True):
        for key in keys:
            assert key in valid_keys, f"{location}: {key} is not a valid parameter under {parameter_name}. " \
                                      f"Valid parameters are: {str(valid_keys)[1:-1]}."

        if exclusive:
            if len(keys) > len(valid_keys):
                raise AssertionError(
                    f"{location}: {str(list(set(keys) - set(valid_keys)))[1:-1]} are not valid parameters "
                    f" under {parameter_name}. Valid parameters are: {str(valid_keys)[1:-1]}. "
                    f"Remove invalid parameters.")
            elif len(keys) < len(valid_keys):
                raise AssertionError(f"{location}: Missing parameters: {str(list(set(valid_keys) - set(keys)))[1:-1]} "
                                     f"under {parameter_name}. Valid parameters are: {str(valid_keys)[1:-1]}. "
                                     f"Please add missing parameters.")

    @staticmethod
    def assert_valid_tabular_file(file_path, location: str, parameter_name: str, sep="\t",
                                  expected_columns: list = None):
        assert Path(file_path).is_file(), f"{location}: {parameter_name} {str(file_path)} is not an existing file."

        if expected_columns is not None:
            columns = pd.read_csv(file_path, index_col=0, nrows=0, sep=sep).columns.tolist()
            assert set(columns) == set(
                expected_columns), f"{location}: columns for {parameter_name} are not as expected.\n" \
                                   f"Expected: {expected_columns}\n" \
                                   f"Found: {columns}"

    @staticmethod
    def assert_sequence_type(params, location: str = ""):
        assert "sequence_type" in params, f"{location}: 'sequence_type' is missing: {params}."
        if not isinstance(params['sequence_type'], SequenceType):
            assert params['sequence_type'].upper() in [st.name for st in SequenceType], \
                (f"{location}: {params['sequence_type']} is not a valid sequence type. "
                 f"Valid sequence types are: {[st.name for st in SequenceType]}.")

    @staticmethod
    def assert_region_type(params, location: str = ""):
        assert "region_type" in params, f"{location}: 'region_type' is missing: {params}."
        if not isinstance(params['region_type'], RegionType):
            assert params['region_type'].upper() in [rt.name for rt in RegionType], \
                (f"{location}: {params['region_type']} is not a valid region type. Valid sequence types "
                 f"are: {[rt.name for rt in RegionType]}.")



# quality: gold
import random
from dataclasses import dataclass
from itertools import combinations
from typing import List

import numpy as np

from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.environment.SequenceType import SequenceType
from immuneML.simulation.implants.Motif import Motif
from immuneML.simulation.implants.MotifInstance import MotifInstance


@dataclass
class SeedMotif(Motif):
    """
    Describes motifs by seed, possible gaps, allowed hamming distances, positions that can be changed and what they can be changed to.

    **Specification arguments:**

    - seed (str): An amino acid sequence that represents the basic motif seed. All implanted motifs correspond to the seed, or a modified version thereof, as specified in its instantiation strategy. If this argument is set, seed_chain1 and seed_chain2 arguments are not used.

    - min_gap (int): The minimum gap length, in case the original seed contains a gap.

    - max_gap (int): The maximum gap length, in case the original seed contains a gap.

    - hamming_distance_probabilities (dict): The probability of modifying the given seed with each number of modifications. The keys represent the number of modifications (hamming distance) between the original seed and the implanted motif, and the values represent the probabilities for the respective number of modifications. For example {0: 0.7, 1: 0.3} means that 30% of the time one position will be modified, and the remaining 70% of the time the motif will remain unmodified with respect to the seed. The values of hamming_distance_probabilities must sum to 1.

    - position_weights (dict): A dictionary containing the relative probabilities of choosing each position for hamming distance modification. The keys represent the position in the seed, where counting starts at 0. If the index of a gap is specified in position_weights, it will be removed. The values represent the relative probabilities for modifying each position when it gets selected for modification. For example {0: 0.6, 1: 0, 2: 0.4} means that when a sequence is selected for a modification (as specified in hamming_distance_probabilities), then 60% of the time the amino acid at index 0 is modified, and the remaining 40% of the time the amino acid at index 2. If the values of position_weights do not sum to 1, the remainder will be redistributed over all positions, including those not specified.

    - alphabet_weights (dict): A dictionary describing the relative probabilities of choosing each amino acid for hamming distance modification. The keys of the dictionary represent the amino acids and the values are the relative probabilities for choosing this amino acid. If the values of alphabet_weights do not sum to 1, the remainder will be redistributed over all possible amino acids, including those not specified.

    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            motifs:
                # examples for single chain receptor data
                my_simple_motif: # this will be the identifier of the motif
                    seed: AAA # motif is always AAA
                my_gapped_motif:
                    seed: AA/A # this motif can be AAA, AA_A, CAA, CA_A, DAA, DA_A, EAA, EA_A
                    min_gap: 0
                    max_gap: 1
                    hamming_distance_probabilities: # it can have a max of 1 substitution
                        0: 0.7
                        1: 0.3
                    position_weights: # note that index 2, the position of the gap, is excluded from position_weights
                        0: 1 # only first position can be changed
                        1: 0
                        3: 0
                    alphabet_weights: # the first A can be replaced by C, D or E
                        C: 0.4
                        D: 0.4
                        E: 0.2

    """

    seed: str = None
    min_gap: int = 0
    max_gap: int = 0
    hamming_distance_probabilities: dict = None
    position_weights: dict = None
    alphabet_weights: dict = None
    all_possible_instances: list = None

    def instantiate_motif(self, sequence_type: SequenceType = SequenceType.AMINO_ACID) -> MotifInstance:
        allowed_positions = list(range(len(self.seed)))
        instance = list(self.seed)

        if "/" in self.seed:
            gap_index = self.seed.index("/")
            allowed_positions.remove(gap_index)

        alphabet_weights = self.set_default_weights(self.alphabet_weights,
                                                    EnvironmentSettings.get_sequence_alphabet(sequence_type=sequence_type))
        position_weights = self.set_default_weights(self.position_weights, allowed_positions)

        gap_size = np.random.choice(range(self.min_gap, self.max_gap + 1))
        instance = self._substitute_letters(position_weights, alphabet_weights, allowed_positions, instance, sequence_type)
        instance = "".join(instance)

        return MotifInstance(instance, gap_size)

    def get_max_length(self):
        return len(self.seed.replace("/", "")) + self.max_gap

    def get_all_possible_instances(self, sequence_type: SequenceType):
        if self.all_possible_instances is None:
            if self.hamming_distance_probabilities:
                self.all_possible_instances = self._get_all_hamming_dist_instances(self.seed, sequence_type)
            else:
                self.all_possible_instances = [self._add_gap(self.seed)]

        return self.all_possible_instances

    def __str__(self):
        return str(vars(self))

    def get_alphabet(self) -> List[str]:
        return [letter for letter in list(self.seed) if letter != "/"]

    def _get_allowed_positions(self, base) -> list:
        allowed_positions = [i for i in range(len(base)) if base[i] != "/"]
        allowed_positions = [key for key, val in self.set_default_weights(self.position_weights, allowed_positions).items() if val > 0]
        return allowed_positions

    def _get_all_motif_regex(self, alphabet_weights: dict, allowed_positions: list, hamming_dist: int, base: str, sequence_type: SequenceType):
        motif_regex_instances = []
        letter_to_use = [letter for letter, weight in alphabet_weights.items() if weight > 0]
        replacement_alphabet = f"[{''.join(letter_to_use)}]" if len(letter_to_use) != len(EnvironmentSettings.get_sequence_alphabet(sequence_type)) else "."

        for position_group in combinations(allowed_positions, hamming_dist):
            motif_parts = [base[i: j] for i, j in zip([0] + [el + 1 for el in position_group], list(position_group) + [len(base)])]
            motif_instance = replacement_alphabet.join(motif_parts)
            motif_instance = self._add_gap(motif_instance)
            motif_regex_instances.append(motif_instance)

        return motif_regex_instances

    def _get_all_hamming_dist_instances(self, base, sequence_type) -> list:
        allowed_positions = self._get_allowed_positions(base)
        alphabet_weights = self.set_default_weights(self.alphabet_weights, EnvironmentSettings.get_sequence_alphabet(sequence_type=sequence_type))
        motif_instances = []

        for hamming_dist, dist_proba in self.hamming_distance_probabilities.items():
            if hamming_dist > 0 and dist_proba > 0:
                motif_regex_instances = self._get_all_motif_regex(alphabet_weights, allowed_positions, hamming_dist, base, sequence_type)
                motif_instances.extend(motif_regex_instances)
            elif dist_proba > 0:
                motif_instance = self._add_gap(base)
                motif_instances.append(motif_instance)

        return motif_instances

    def _add_gap(self, motif_instance):
        if self.max_gap > 0 or "/" in motif_instance:
            gap_length = f"{self.min_gap},{self.max_gap}"
            return motif_instance.replace("/", ".{" + gap_length + "}")
        else:
            return motif_instance

    def _substitute_letters(self, position_weights, alphabet_weights, allowed_positions: list, instance: list,
                            sequence_type: SequenceType):

        if self.hamming_distance_probabilities:
            substitution_count = random.choices(list(self.hamming_distance_probabilities.keys()),
                                                list(self.hamming_distance_probabilities.values()), k=1)[0]
            allowed_position_weights = {key: value for key, value in position_weights.items() if key in allowed_positions}
            position_probabilities = self._prepare_probabilities(allowed_position_weights)
            positions = list(np.random.choice(allowed_positions, size=substitution_count, p=position_probabilities))

            while substitution_count > 0:
                if position_weights[positions[substitution_count - 1]] > 0:  # if the position is allowed to be changed
                    position = positions[substitution_count - 1]
                    alphabet_probabilities = self._prepare_probabilities(alphabet_weights)
                    instance[position] = np.random.choice(EnvironmentSettings.get_sequence_alphabet(sequence_type),
                                                          size=1, p=alphabet_probabilities)[0]
                substitution_count -= 1

        return instance

    def _prepare_keys(self, weights):
        keys = list(weights.keys())
        keys.sort()
        return keys

    def _prepare_probabilities(self, weights: dict):
        keys = self._prepare_keys(weights)
        s = sum([weights[key] for key in keys])
        return [weights[key] / s for key in keys]

    def set_default_weights(self, weights, allowed_keys):
        weights = {k: v for k, v in weights.items() if k in allowed_keys} if weights is not None else {}
        weight_sum = sum(list(weights.values()))

        if 0.99 <= weight_sum <= 1.:
            weights = {**{key: 0 for key in allowed_keys}, **weights}
        else:
            missing_keys = [key for key in allowed_keys if key not in weights]
            weights = {**{key: (1 - weight_sum) / len(missing_keys) for key in missing_keys}, **weights}
        return weights


import numpy as np
import re
from immuneML.data_model.AIRRSequenceSet import AIRRSequenceSet
from immuneML.data_model.SequenceParams import RegionType
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.encodings.protein_embedding.ProteinEmbeddingEncoder import ProteinEmbeddingEncoder
from immuneML.util.ParameterValidator import ParameterValidator


class ProtT5Encoder(ProteinEmbeddingEncoder):

    """
        Encoder based on a pretrained protein language model by Elnaggar et al. 2021. The used transformer model is
        "Rostlab/prot_t5_xl_half_uniref50-enc".

        Original publication:
        Elnaggar, A., Heinzinger, M., Dallago, C., Rihawi, G., Wang, Y., Jones, L., Gibbs, T., Feher, T.,
        Angerer, C., Steinegger, M., Bhowmik, D., & Rost, B. (2021). ProtTrans: Towards Cracking the Language of
        Life's Code Through Self-Supervised Deep Learning and High Performance Computing (No. arXiv:2007.06225).
        arXiv. https://doi.org/10.48550/arXiv.2007.06225

        Original GitHub repository with license information: https://github.com/agemagician/ProtTrans

        **Specification arguments:**

        - region_type (RegionType): Which part of the receptor sequence to encode. Defaults to IMGT_CDR3.

        - device (str): Which device to use for model inference - 'cpu', 'cuda', 'mps' - as defined by pytorch.
          Defaults to 'cpu'.

        - num_processes (int): Number of processes to use for parallel processing. Defaults to 1.

        """

    def __init__(self, name: str = None, region_type: RegionType = RegionType.IMGT_CDR3, device: str = 'cpu',
                 num_processes: int = 1):
        super().__init__(region_type, name)
        self.region_type = region_type
        self.device = device
        self.transformer_link = "Rostlab/prot_t5_xl_half_uniref50-enc"
        self.num_processes = num_processes

    @staticmethod
    def build_object(dataset: Dataset, **params):
        ParameterValidator.assert_region_type(params, ProtT5Encoder.__name__)
        return ProtT5Encoder(**{**params, 'region_type': RegionType[params['region_type'].upper()]})

    def _embed_sequence_set(self, sequence_set: AIRRSequenceSet, seq_field: str):
        import torch
        from transformers import T5Tokenizer, T5EncoderModel

        print("Loading: {}".format(self.transformer_link))
        model = T5EncoderModel.from_pretrained(self.transformer_link)
        if self.device == torch.device("cpu"):
            print("Casting model to full precision for running on CPU ...")
            model.to(torch.float32)  # only cast to full-precision if no GPU is available
        model = model.to(self.device)
        model = model.eval()
        tokenizer = T5Tokenizer.from_pretrained(self.transformer_link, do_lower_case=False, legacy=True)

        sequences = getattr(sequence_set, seq_field)
        sequence_lengths = sequences.lengths
        sequences = sequences.tolist()
        sequences = [" ".join(list(re.sub(r"[UZOB]", "X", sequence))) for sequence in sequences]

        ids = tokenizer.batch_encode_plus(sequences, add_special_tokens=True, padding="longest")
        input_ids = torch.tensor(ids['input_ids']).to(self.device)
        attention_mask = torch.tensor(ids['attention_mask']).to(self.device)

        with torch.no_grad():
            embedding_repr = model(input_ids=input_ids, attention_mask=attention_mask)

        embeddings = [embedding_repr.last_hidden_state[i, :sequence_lengths[i]].mean(dim=0).numpy(force=True)
                      for i in range(len(sequences))]

        return np.array(embeddings)

    def _get_encoding_name(self) -> str:
        return f"ProtT5Encoder({self.transformer_link})"

    def _get_model_link(self) -> str:
        return self.transformer_link

    def _get_caching_params(self, dataset, params: EncoderParams):
        cache_params = (dataset.identifier, self.__class__.__name__, self.region_type.name, self._get_model_link())
        return cache_params



from immuneML.IO.dataset_import.DataImport import DataImport
from immuneML.data_model.SequenceParams import ChainPair, RegionType
from scripts.specification_util import update_docs_per_mapping


class AIRRImport(DataImport):
    """
    Imports data in AIRR format into a Repertoire-, Sequence- or ReceptorDataset.
    RepertoireDatasets should be used when making predictions per repertoire, such as predicting a disease state.
    SequenceDatasets or ReceptorDatasets should be used when predicting values for unpaired (single-chain) and paired
    immune receptors respectively, like antigen specificity.

    The AIRR .tsv format is explained here: https://docs.airr-community.org/en/stable/datarep/format.html
    And the AIRR rearrangement schema can be found here: https://docs.airr-community.org/en/stable/datarep/rearrangements.html

    When importing a ReceptorDataset, the AIRR field cell_id is used to determine the chain pairs.

    **Specification arguments:**

    - path (str): For RepertoireDatasets, this is the path to a directory with AIRR files to import. For Sequence- or ReceptorDatasets this path may either be the path to the file to import, or the path to the folder locating one or multiple files with .tsv, .csv or .txt extensions. By default path is set to the current working directory.

    - is_repertoire (bool): If True, this imports a RepertoireDataset. If False, it imports a SequenceDataset or ReceptorDataset. By default, is_repertoire is set to True.

    - metadata_file (str): Required for RepertoireDatasets. This parameter specifies the path to the metadata file. This is a csv file with columns filename, subject_id and arbitrary other columns which can be used as labels in instructions. Only the AIRR files included under the column 'filename' are imported into the RepertoireDataset. For setting Sequence- or ReceptorDataset labels, metadata_file is ignored, use label_columns instead.

    - label_columns (list): For Sequence- or ReceptorDataset, this parameter can be used to explicitly set the column names of labels to import. These labels can be used as prediction target. When label_columns are not set, label names are attempted to be discovered automatically (any column name which is not used in the column_mapping). For setting RepertoireDataset labels, label_columns is ignored, use metadata_file instead.

    - paired (str): Required for Sequence- or ReceptorDatasets. This parameter determines whether to import a SequenceDataset (paired = False) or a ReceptorDataset (paired = True). In a ReceptorDataset, two sequences with chain types specified by receptor_chains are paired together based on the identifier given in the AIRR column named 'cell_id'.

    - receptor_chains (str): Required for ReceptorDatasets. Determines which pair of chains to import for each Receptor. Valid values for receptor_chains are the names of the :py:obj:`~immuneML.data_model.receptor.ChainPair.ChainPair` enum. If receptor_chains is not provided, the chain pair is automatically detected (only one chain pair type allowed per repertoire).

    - import_productive (bool): Whether productive sequences (with value 'T' in column productive) should be included in the imported sequences. By default, import_productive is True.

    - import_unknown_productivity (bool): Whether sequences with unknown productivity (missing value in column productive) should be included in the imported sequences. By default, import_unknown_productivity is True.

    - import_with_stop_codon (bool): Whether sequences with stop codons (with value 'T' in column stop_codon) should be included in the imported sequences. This only applies if column stop_codon is present. By default, import_with_stop_codon is False.

    - import_out_of_frame (bool): Whether out of frame sequences (with value 'F' in column vj_in_frame) should be included in the imported sequences. This only applies if column vj_in_frame is present. By default, import_out_of_frame is False.

    - import_illegal_characters (bool): Whether to import sequences that contain illegal characters, i.e., characters that do not appear in the sequence alphabet (amino acids including stop codon '*', or nucleotides). When set to false, filtering is only applied to the sequence type of interest (when running immuneML in amino acid mode, only entries with illegal characters in the amino acid sequence are removed). By default import_illegal_characters is False.

    - import_empty_nt_sequences (bool): imports sequences which have an empty nucleotide sequence field; can be True or False. By default, import_empty_nt_sequences is set to True.

    - import_empty_aa_sequences (bool): imports sequences which have an empty amino acid sequence field; can be True or False; for analysis on amino acid sequences, this parameter should be False (import only non-empty amino acid sequences). By default, import_empty_aa_sequences is set to False.

    - column_mapping (dict): A mapping from AIRR column names to immuneML's internal data representation. A custom column mapping can be specified here if necessary (for example; adding additional data fields if they are present in the AIRR file, or using alternative column names).

        .. indent with spaces
        .. code-block:: yaml

            additional_column_in_the_file: column_name_to_be_used_in_analysis

    - separator (str): Column separator, for AIRR this is by default "\\t".


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            datasets:
                my_airr_dataset:
                    format: AIRR
                    params:
                        path: path/to/files/
                        is_repertoire: True # whether to import a RepertoireDataset
                        metadata_file: path/to/metadata.csv # metadata file for RepertoireDataset
                        import_productive: True # whether to include productive sequences in the dataset
                        import_with_stop_codon: False # whether to include sequences with stop codon in the dataset
                        import_out_of_frame: False # whether to include out of frame sequences in the dataset
                        import_illegal_characters: False # remove sequences with illegal characters for the sequence_type being used
                        import_empty_nt_sequences: True # keep sequences even if the `sequences` column is empty (provided that other fields are as specified here)
                        import_empty_aa_sequences: False # remove all sequences with empty column
                        # Optional fields with AIRR-specific defaults, only change when different behavior is required:
                        separator: "\\t" # column separator
                        region_type: IMGT_CDR3 # what part of the sequence check for import

    """

    @staticmethod
    def get_documentation():
        doc = str(AIRRImport.__doc__)

        chain_pair_values = str([chain_pair.name for chain_pair in ChainPair])[1:-1].replace("'", "`")
        region_type_values = str([region_type.name for region_type in RegionType])[1:-1].replace("'", "`")

        mapping = {
            "Valid values for receptor_chains are the names of the :py:obj:`~immuneML.data_model.receptor.ChainPair.ChainPair` enum.": f"Valid values are {chain_pair_values}.",
            "Valid values for region_type are the names of the :py:obj:`~immuneML.data_model.receptor.RegionType.RegionType` enum.": f"Valid values are {region_type_values}.",
        }
        doc = update_docs_per_mapping(doc, mapping)
        return doc



import datetime
import logging


def log(func):
    def wrapped(*args, **kwargs):
        try:
            logging.info("--- Entering: %s with parameters %s" % (func.__name__, args))
            try:
                return func(*args, **kwargs)
            except Exception as e:
                logging.error('\n\n--- Exception in %s : %s\n\n' % (func.__name__, e))
                if "dsl" in func.__globals__["__name__"]:
                    raise Exception(f"{e}\n\n"
                                    f"ImmuneMLParser: an error occurred during parsing in function {func.__name__} "
                                    f" with parameters: {args}.\n\nFor more details on how to write the specification, "
                                    f"see the documentation. For technical description of the error, see the log above.").with_traceback(e.__traceback__)
                else:
                    raise e
        finally:
            logging.info("--- Exiting: %s" % (func.__name__))

    return wrapped


def print_log(mssg, include_datetime=False, log_func_name='info'):
    getattr(logging, log_func_name)(mssg)

    if include_datetime:
        mssg = f"{datetime.datetime.now()}: {mssg}"

    print(mssg, flush=True)



class HPLabelState:

    def __init__(self, label, auxiliary_labels):
        self.label = label
        self.auxiliary_labels = auxiliary_labels

        # computed
        self.assessment_items = {}
        self.selection_state = None

    @property
    def optimal_assessment_item(self):
        return self.assessment_items[str(self.selection_state.optimal_hp_setting)]

    @property
    def optimal_hp_setting(self):
        return self.selection_state.optimal_hp_setting



from enum import Enum


class ClassificationMetric(Enum):

    ACCURACY = "accuracy_score"
    BALANCED_ACCURACY = "balanced_accuracy_score"
    CONFUSION_MATRIX = "confusion_matrix"
    F1_MICRO = "f1_score_micro"
    F1_MACRO = "f1_score_macro"
    F1_WEIGHTED = "f1_score_weighted"
    PRECISION = "precision_score"
    RECALL = "recall_score"
    AUC = "roc_auc_score"
    LOG_LOSS = "log_loss"

    @staticmethod
    def get_metric(metric_name: str):
        try:
            return ClassificationMetric[metric_name.upper()]
        except KeyError as e:
            raise KeyError(f"'{metric_name}' is not a valid performance metric. Valid metrics are: {', '.join([m.name for m in ClassificationMetric])}").with_traceback(e.__traceback__)

    @staticmethod
    def get_search_criterion(metric):
        if metric in [ClassificationMetric.LOG_LOSS]:
            return min
        else:
            return max

    @staticmethod
    def get_sklearn_score_name(metric):
        if metric in [ClassificationMetric.LOG_LOSS]:
            return f"neg_{metric.name.lower()}"
        else:
            return metric.name.lower()

    @staticmethod
    def get_probability_based_metric_types():
        return [ClassificationMetric.LOG_LOSS, ClassificationMetric.AUC]


import logging
import os
import pickle
from pathlib import Path

import numpy as np

from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.encodings.abundance_encoding.AbundanceEncoderHelper import AbundanceEncoderHelper
from immuneML.pairwise_repertoire_comparison.ComparisonDataBatch import ComparisonDataBatch
from immuneML.util.Logger import log
from immuneML.util.PathBuilder import PathBuilder


class ComparisonData:

    @log
    def __init__(self, repertoire_ids: list, comparison_attributes, sequence_batch_size: int = 10000, path: Path = None):

        self.path = PathBuilder.build(path / "comparison_data")
        self.sequence_batch_size = sequence_batch_size
        self.item_count = 0
        self.comparison_attributes = comparison_attributes
        self.repertoire_ids = repertoire_ids
        self.iteration_repertoire_ids = None
        self.batches = []
        self.tmp_batch_paths = []

    def __len__(self):
        return self.item_count

    def __iter__(self):
        return self.get_item_vectors(self.iteration_repertoire_ids)

    def set_iteration_repertoire_ids(self, iteration_repertoire_ids):
        """Defines the subset and order of repertoires to iterate over"""
        self.iteration_repertoire_ids = iteration_repertoire_ids

    def get_item_names(self):
        return np.array([item for items in [batch.get_items() for batch in self.batches] for item in items])

    def get_item_vectors(self, repertoire_ids: list = None):
        for batch in self.get_batches(columns=repertoire_ids):
            for item_index in range(batch.shape[0]):
                yield batch[item_index]

    def get_repertoire_vectors(self, identifiers: list):
        repertoire_vectors = {identifier: np.zeros(self.item_count) for identifier in identifiers}
        for batch_index, batch in enumerate(self.get_batches(columns=identifiers, return_dict=True)):
            start = batch_index * self.sequence_batch_size
            for identifier in identifiers:
                end = start + batch[identifier].shape[0]
                repertoire_vectors[identifier][start: end] = batch[identifier]
        return repertoire_vectors

    def get_repertoire_vector(self, identifier: str):
        repertoire_vector = np.zeros(self.item_count)
        for batch_index, batch in enumerate(self.get_batches(columns=[identifier])):
            start = batch_index * self.sequence_batch_size
            end = start + batch.shape[0]
            repertoire_vector[start: end] = batch[:, 0]
        return repertoire_vector

    def get_item_vector(self, index: int):
        batch_index = int(index / self.sequence_batch_size)
        index_in_batch = index - (batch_index * self.sequence_batch_size)
        return self.batches[batch_index].get_matrix()[index_in_batch]

    def get_batches(self, columns: list = None, return_dict: bool = False):
        for index in range(len(self.batches)):
            yield self.get_batch(index, columns, return_dict)

    def get_batch(self, index: int, columns: list = None, return_dict: bool = False):
        batch = self.batches[index].load()
        if columns is not None and not return_dict:
            column_indices = [self.batches[index].repertoire_index_mapping[col] for col in columns]
            return batch.get_matrix()[:, column_indices]
        elif columns is not None:
            column_indices = [self.batches[index].repertoire_index_mapping[col] for col in columns]
            matrix = batch.get_matrix()
            return {col: matrix[:, column_indices[i]] for i, col in enumerate(columns)}
        else:
            return batch.get_matrix()

    @log
    def process_dataset(self, dataset: RepertoireDataset):
        extract_fn = AbundanceEncoderHelper.get_matching_func_for_repertoire(self.comparison_attributes)
        repertoire_count = dataset.get_example_count()
        for index, repertoire in enumerate(dataset.get_data()):
            self.process_repertoire(repertoire, str(repertoire.identifier), extract_fn)
            logging.info("Repertoire {} ({}/{}) processed.".format(repertoire.identifier, index + 1, repertoire_count))
            logging.info(f"Currently, there are {self.item_count} items in the comparison data matrix.")
        self.merge_tmp_batches_to_matrix()

    def merge_tmp_batches_to_matrix(self):

        for index in range(len(self.tmp_batch_paths)):

            batch = self.load_tmp_batch(index)
            matrix = np.zeros((self.sequence_batch_size, len(self.repertoire_ids)), order='F')
            items = []

            for item_index, item in enumerate(batch):
                items.append(item)
                for repertoire_index, repertoire_id in enumerate(self.repertoire_ids):
                    if repertoire_id in batch[item]:
                        matrix[item_index][repertoire_index] = batch[item][repertoire_id]

            df = np.array(matrix[:len(items)], order='F')
            repertoire_index_mapping = {rep_id: ind for ind, rep_id in enumerate(self.repertoire_ids)}

            comp_data_batch = ComparisonDataBatch(matrix=df, items=items, repertoire_index_mapping=repertoire_index_mapping, path=self.path,
                                                  identifier=index)
            comp_data_batch.store()

            self.batches.append(comp_data_batch)

        for path in self.tmp_batch_paths:
            os.remove(path)

    @log
    def process_repertoire(self, repertoire, repertoire_id: str, extract_items_fn):
        items = extract_items_fn(repertoire)
        new_items = self.filter_existing_items(items, repertoire_id)
        self.add_items_for_repertoire(new_items, repertoire_id)

    def filter_existing_items(self, items: list, repertoire_id: str) -> list:
        new_items = items
        for batch_index, batch in enumerate(self.get_tmp_batches()):
            new_items = self._remove_existing_items(new_items, batch, batch_index, repertoire_id)
        return new_items

    def _match_items_to_batch(self, items, batch):

        keep = set(items).difference(batch)

        item_to_update = set(items).intersection(batch)
        value = np.ones(len(item_to_update), dtype=np.bool_)

        return keep, value, item_to_update

    def _remove_existing_items(self, new_items: list, batch: dict, batch_index: int, repertoire_id: str) -> list:

        update = {"value": [], "index": []}

        new_items_to_keep, update["value"], update["index"] = self._match_items_to_batch(new_items, batch)

        for index, item in enumerate(update["index"]):
            batch[item][repertoire_id] = update["value"][index]

        self.store_tmp_batch(batch, batch_index)

        return list(new_items_to_keep)

    def store_tmp_batch(self, batch: dict, batch_index: int):

        if len(self.tmp_batch_paths) > batch_index or len(self.tmp_batch_paths) == batch_index:
            batch_path = self.path / f'tmp_batch_{batch_index}.pickle'
            with batch_path.open('wb') as file:
                pickle.dump(batch, file)
            if len(self.tmp_batch_paths) == batch_index:
                self.tmp_batch_paths.append(batch_path)
        else:
            raise KeyError("ComparisonData: batch_index: {} does not exist. tmp_batches length: {}"
                           .format(batch_index, len(self.tmp_batch_paths)))

    def get_tmp_batches(self):
        for i in range(len(self.tmp_batch_paths)):
            yield self.load_tmp_batch(i)

    def load_tmp_batch(self, batch_index: int) -> dict:
        if len(self.tmp_batch_paths) > 0 and self.tmp_batch_paths[batch_index].is_file():
            with self.tmp_batch_paths[batch_index].open("rb") as file:
                batch = pickle.load(file)
        else:
            batch = {}
        return batch

    def add_items_for_repertoire(self, items: list, repertoire_id: str):
        last_batch_index = len(self.tmp_batch_paths)-1 if len(self.tmp_batch_paths) > 0 else 0
        batch = self.load_tmp_batch(last_batch_index)

        self.item_count += len(items)

        item_index = 0
        while len(batch) < self.sequence_batch_size and item_index < len(items):
            batch[items[item_index]] = {repertoire_id: 1}
            item_index += 1

        self.store_tmp_batch(batch, last_batch_index)

        items = items[item_index:]

        while len(items) > 0:
            batch = {item: {repertoire_id: 1} for item in items[:self.sequence_batch_size]}
            last_batch_index += 1
            self.store_tmp_batch(batch, last_batch_index)
            items = items[self.sequence_batch_size:]



import logging
from pathlib import Path
from typing import Tuple, List

import numpy as np
import pandas as pd
import plotly.graph_objects as go

from immuneML.environment.Label import Label
from immuneML.hyperparameter_optimization.states.TrainMLModelState import TrainMLModelState
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.train_ml_model_reports.TrainMLModelReport import TrainMLModelReport
from immuneML.util.PathBuilder import PathBuilder


class CVFeaturePerformance(TrainMLModelReport):
    """
    This report plots the average training vs test performance w.r.t. given encoding parameter which is explicitly set
    in the feature attribute. It can be used only in combination with TrainMLModel instruction and can be only specified under 'reports'


    **Specification arguments:**

    - feature: name of the encoder parameter w.r.t. which the performance across training and test will be shown. Possible values depend
      on the encoder on which it is used.

    - is_feature_axis_categorical (bool): if the x-axis of the plot where features are shown should be categorical; alternatively it is
      automatically determined based on the feature values


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            reports:
                report1:
                    CVFeaturePerformance:
                        feature: p_value_threshold # parameter value of SequenceAbundance encoder
                        is_feature_axis_categorical: True # show x-axis as categorical

    """

    @classmethod
    def build_object(cls, **kwargs):
        return CVFeaturePerformance(**kwargs)

    def __init__(self, feature: str = None, state: TrainMLModelState = None, result_path: Path = None, label: Label = None,
                 name: str = None, is_feature_axis_categorical: bool = None, number_of_processes: int = 1):
        super().__init__(name=name, state=state, label=label, result_path=result_path, number_of_processes=number_of_processes)
        self.feature = feature
        self.relevant_hp_settings = []
        self.feature_values = None
        self.feature_count = None
        self.result_name = None
        self.is_feature_axis_categorical = is_feature_axis_categorical

    def check_prerequisites(self):
        self._extract_label()

        if self.label is None:
            logging.warning("CVFeaturePerformance: the label was not set for this report and it could not be inferred from the instruction "
                          "as there might be multiple labels there. Skipping the report.", RuntimeWarning)
            return False

        self._extract_hp_settings()
        if self.feature_count != len(self.relevant_hp_settings):
            logging.warning(f"CVFeaturePerformance: there are multiple hyperparameter settings with the same value of the "
                          f"feature {self.feature}. Skipping the report...", RuntimeWarning)
            return False

        return True

    def _extract_label(self):
        if self.label is None and len(self.state.label_configuration.get_label_objects()) == 1:
            self.label = self.state.label_configuration.get_label_objects()[0]

    def _extract_hp_settings(self):
        self.relevant_hp_settings = [hp_setting for hp_setting in self.state.hp_settings
                                     if self.feature in hp_setting.encoder_params]
        self.feature_values = np.unique([hp_setting.encoder_params[self.feature] for hp_setting in self.relevant_hp_settings])
        self.feature_count = len(self.feature_values)

    def _generate(self) -> ReportResult:

        PathBuilder.build(self.result_path)
        self.result_name = f"{self.feature}_performance"

        training_dataframe, test_dataframe = self._make_plot_dataframes()
        table_results = self._store_dataframes(training_dataframe, test_dataframe)

        report_output_fig = self._plot(training_dataframe=training_dataframe, test_dataframe=test_dataframe)
        output_figures = None if report_output_fig is None else [report_output_fig]

        return ReportResult(name=self.name,
                            info=f"The average training vs test performance w.r.t. the given encoding parameter {self.feature}.",
                            output_tables=table_results,
                            output_figures=output_figures)

    def _plot(self, training_dataframe, test_dataframe):

        optimization_metric = self.state.optimization_metric.name.lower()

        fig = go.Figure()
        fig.add_trace(go.Scatter(x=training_dataframe["x"], y=training_dataframe["y"], name="training", mode="markers", marker_size=11, marker_color="#CC79A7",
                                 hovertemplate=f"training {optimization_metric}" + ": %{y}<extra></extra>", opacity=0.8))
        fig.add_trace(go.Scatter(x=test_dataframe["x"], y=test_dataframe["y"], name="test", mode="markers", marker_size=11, marker_color="#009E73",
                                 hovertemplate=f"test {optimization_metric}" + ": %{y}<extra></extra>", opacity=0.8))
        fig.update_layout(legend_title_text="Data", title="Performance across feature values", template="plotly_white")
        fig.update_xaxes(title_text=self.feature)
        if self.is_feature_axis_categorical:
            fig.update_xaxes(type='category')
        fig.update_yaxes(title_text=f"performance ({self.state.optimization_metric.name.lower()})")
        fig.update_layout(hovermode="x unified")

        file_path = self.result_path / f"{self.result_name}.html"
        fig.write_html(str(file_path))

        return ReportOutput(path=file_path)

    def _store_dataframes(self, training_dataframe: pd.DataFrame, test_dataframe: pd.DataFrame) -> List[ReportOutput]:
        train_path = self.result_path / "training_performance.csv"
        test_path = self.result_path / "test_performance.csv"
        training_dataframe.to_csv(train_path, index=False)
        test_dataframe.to_csv(test_path, index=False)

        return [ReportOutput(path=train_path, name=f"Training performance w.r.t. {self.feature} values"),
                ReportOutput(path=test_path, name=f"Test performance w.r.t. {self.feature} values")]

    def _make_plot_dataframes(self) -> Tuple[pd.DataFrame, pd.DataFrame]:

        performance_training = np.zeros((self.feature_count, self.state.assessment.split_count,
                                         self.state.selection.split_count))
        features_test = np.zeros((self.state.assessment.split_count, self.feature_count))
        performance_test = np.zeros((self.state.assessment.split_count, self.feature_count))

        for assessment_split_index, assessment_state in enumerate(self.state.assessment_states):

            assessment_items = [assessment_state.label_states[self.label.name].assessment_items[hp_setting.get_key()]
                                for hp_setting in self.relevant_hp_settings]
            features_test[assessment_split_index] = [item.hp_setting.encoder_params[self.feature] for item in assessment_items]
            performance_test[assessment_split_index] = [item.performance[self.state.optimization_metric.name.lower()] for item in assessment_items]

            for hp_index, hp_setting in enumerate(self.relevant_hp_settings):
                performance_training[hp_index, assessment_split_index] = \
                    [item.performance[self.state.optimization_metric.name.lower()]
                     for item in assessment_state.label_states[self.label.name].selection_state.hp_items[hp_setting.get_key()]]

        feature_values = self.feature_values.astype(str)

        test_dataframe = pd.DataFrame({"x": feature_values, "y": performance_test.mean(axis=0)})
        training_dataframe = pd.DataFrame({"x": feature_values, "y": performance_training.mean(axis=(1, 2))})

        return training_dataframe, test_dataframe



import random
from dataclasses import dataclass
from pathlib import Path
from typing import List

import numpy as np
from bionumpy.io.motifs import read_motif
from bionumpy.sequence.position_weight_matrix import PWM as bnp_PWM

from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.environment.SequenceType import SequenceType
from immuneML.simulation.implants.Motif import Motif
from immuneML.simulation.implants.MotifInstance import MotifInstance


@dataclass
class LigoPWM(Motif):
    """
    Motifs defined by a positional weight matrix and using bionumpy's PWM internally.
    For more details on bionumpy's implementation of PWM, as well as for supported formats,
    see the documentation at https://bionumpy.github.io/bionumpy/tutorials/position_weight_matrix.html.

    **Specification arguments:**

    - file_path: path to the file where the PWM is stored

    - threshold (float): when matching PWM to a sequence, this is the threshold to consider the sequence as containing the motif

    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            motifs:
                my_custom_pwm: # this will be the identifier of the motif
                    file_path: my_pwm_1.csv
                    threshold: 2

    """
    file_path: Path
    pwm_matrix: bnp_PWM  # with log-likelihood
    threshold: float

    @classmethod
    def build(cls, identifier: str, file_path, threshold: float):
        assert Path(file_path).is_file(), file_path
        pwm_matrix = read_motif(file_path)
        return LigoPWM(identifier, file_path, pwm_matrix, threshold)

    def get_all_possible_instances(self, sequence_type: SequenceType):
        assert sorted(self.pwm_matrix.alphabet) == sorted(EnvironmentSettings.get_sequence_alphabet(sequence_type))
        return self

    def get_max_length(self) -> int:
        return self.pwm_matrix.window_size

    def get_alphabet(self) -> List[str]:
        return list(self.pwm_matrix.alphabet)

    def instantiate_motif(self, sequence_type: SequenceType = SequenceType.AMINO_ACID):
        if EnvironmentSettings.get_sequence_alphabet(sequence_type) != sorted(list(self.pwm_matrix.alphabet)):
            raise RuntimeError(
                f"{LigoPWM.__name__}: could not instantiate motif for sequence type {sequence_type.name},"
                f" check if the motif sequence type is a match at {self.file_path}.")

        counts_per_position = np.exp(self.pwm_matrix._matrix + np.log([0.25])[:, np.newaxis])

        return MotifInstance(
            "".join([random.choices(list(self.pwm_matrix.alphabet), weights=counts_per_position[:, position])[0]
                     for position in range(self.pwm_matrix.window_size)]), gap=0)


from pathlib import Path
import pandas as pd

from immuneML.example_weighting.ExampleWeightingParams import ExampleWeightingParams
from immuneML.example_weighting.ExampleWeightingStrategy import ExampleWeightingStrategy
from immuneML.util.ParameterValidator import ParameterValidator


class PredefinedWeighting(ExampleWeightingStrategy):
    '''

    Example weighting strategy where weights are supplied in a file.

    **Specification arguments:**

    - file_path (Path): Path to the example weights, should contain the columns 'identifier' and 'example_weight':

      ==========  ==============
      identifier  example_weight
      ==========  ==============
      1           0.5
      2           1
      3           1
      ========  ==============

    - separator (str): Column separator in the input file.

    '''

    def __init__(self, file_path, separator, name: str = None):
        super().__init__(name)
        self.file_path = Path(file_path)
        self.separator = separator

    @staticmethod
    def _prepare_parameters(file_path, separator, name: str = None):
        file_path = Path(file_path)

        if not file_path.is_file():
            raise FileNotFoundError(f"{PredefinedWeighting.__class__.__name__}: example weigths could not be loaded from {file_path}. "
                                    f"Check if the path to the file is properly set.")

        ParameterValidator.assert_type_and_value(separator, str, location=PredefinedWeighting.__name__, parameter_name="separator")

        return {
            "file_path": file_path,
            "separator": separator,
            "name": name
        }

    @staticmethod
    def build_object(dataset=None, **params):
        prepared_params = PredefinedWeighting._prepare_parameters(**params)

        return PredefinedWeighting(**prepared_params)

    def compute_weights(self, dataset, params: ExampleWeightingParams):
        weights_df = self._read_example_weights_file()

        return self._get_example_weights(dataset, weights_df)

    def _get_example_weights(self, dataset, weights_df):
        return [self._get_example_weight_by_identifier(example.identifier, weights_df) for example in dataset.get_data()]

    def _read_example_weights_file(self):
        return pd.read_csv(self.file_path, sep=self.separator, usecols=["identifier", "example_weight"])

    def _get_example_weight_by_identifier(self, identifier, weights_df):
        return float(weights_df[weights_df["identifier"] == identifier].example_weight)



import abc
from typing import List

from immuneML.data_model.datasets.Dataset import Dataset


class DimRedMethod:
    """

    Dimensionality reduction methods are algorithms which can be used to reduce the dimensionality
    of encoded datasets, in order to uncover and analyze patterns present in the data.

    These methods can be used in the :ref:`ExploratoryAnalysis` and :ref:`Clustering` instructions.
    """

    DOCS_TITLE = "Dimensionality reduction methods"

    def __init__(self, name: str = None):
        self.method = None
        self.name = name

    @abc.abstractmethod
    def fit(self, dataset: Dataset):
        pass

    @abc.abstractmethod
    def transform(self, dataset: Dataset):
        pass

    @abc.abstractmethod
    def fit_transform(self, dataset: Dataset):
        pass

    @abc.abstractmethod
    def get_dimension_names(self) -> List[str]:
        pass

    @classmethod
    def get_title(cls):
        return "Dimensionality Reduction"



import logging
import os
import pathlib
import shutil
import sys
import warnings
from pathlib import Path

import yaml

from immuneML.app.ImmuneMLApp import ImmuneMLApp
from immuneML.util.PathBuilder import PathBuilder


class Quickstart:

    def create_specfication(self, path: Path):

        specs = {
            "definitions": {
                "datasets": {
                    "d1": {
                        "format": "AIRR",
                        "params": {
                            'path': f"{path}/../synthetic_dataset/result/simulation_instruction/exported_dataset/airr",
                            "dataset_file": f"{path}/../synthetic_dataset/result/simulation_instruction/exported_dataset/airr/dataset.yaml",
                            'metadata_file': f"{path}/../synthetic_dataset/result/simulation_instruction/exported_dataset/airr/metadata.csv",
                            "import_illegal_characters": True,
                            'import_productive': False,
                            'import_out_of_frame': True
                        }
                    }
                },
                "encodings": {
                    "e1": {
                        "KmerFrequency": {
                            "k": 3
                        }
                    },
                    "e2": {
                        "KmerFrequency": {
                            "k": 2
                        }
                    }
                },
                "ml_methods": {
                    "simpleLR": {
                        "LogisticRegression": {
                            "C": 0.1,
                            "penalty": "l1",
                            "max_iter": 200
                        }}
                },
                "reports": {
                    "rep1": {
                        "SequenceLengthDistribution": {
                            "batch_size": 3,
                            "region_type": "IMGT_CDR3",
                            "sequence_type": "amino_acid"
                        }
                    },
                    "hprep": "MLSettingsPerformance",
                    "coef": "Coefficients"
                },
            },
            "instructions": {
                "machine_learning_instruction": {
                    "type": "TrainMLModel",
                    "settings": [
                        {
                            "encoding": "e1",
                            "ml_method": "simpleLR"
                        },
                        {
                            "encoding": "e2",
                            "ml_method": "simpleLR"
                        }
                    ],
                    "assessment": {
                        "split_strategy": "random",
                        "split_count": 1,
                        "training_percentage": 0.7,
                        "reports": {
                            "data_splits": ["rep1"],
                            'models': ["coef"]
                        }
                    },
                    "selection": {
                        "split_strategy": "random",
                        "split_count": 1,
                        "training_percentage": 0.7,
                        "reports": {
                            "data_splits": ["rep1"],
                            "models": [],
                        }
                    },
                    "labels": ["my_signal"],
                    "dataset": "d1",
                    "strategy": "GridSearch",
                    "metrics": ["accuracy"],
                    "reports": ["hprep"],
                    "number_of_processes": 3,
                    "optimization_metric": "balanced_accuracy",
                    "refit_optimal_model": False,
                    'region_type': 'IMGT_CDR3',
                    'sequence_type': "amino_acid"
                }
            }
        }
        PathBuilder.build(path)
        specs_file = path / "specs.yaml"
        with specs_file.open("w") as file:
            yaml.dump(specs, file)

        return specs_file

    def build_path(self, path: str = None):
        if path is None:
            path = pathlib.Path.cwd() / "quickstart/"
            if os.path.isdir(path):
                shutil.rmtree(path)
            PathBuilder.build(path)
        else:
            path = PathBuilder.build(path)
        return path

    def _simulate_dataset_with_signals(self, path: Path):

        print("immuneML quickstart: generating a synthetic dataset...")

        PathBuilder.build(path)

        specs = {
            "definitions": {
                "motifs": {"my_motif": {"seed": "AAA"}},
                "signals": {"my_signal": {"motifs": ["my_motif"]}},
                "simulations": {
                    "quickstart_simulation": {
                        'is_repertoire': True,
                        'paired': False,
                        'sequence_type': 'amino_acid',
                        'simulation_strategy': 'Implanting',
                        'remove_seqs_with_signals': False,
                        'sim_items': {
                            "pos_reps": {
                                "signals": {"my_signal": 0.2},
                                "number_of_examples": 50,
                                "receptors_in_repertoire_count": 10,
                                "generative_model": {
                                    "default_model_name": "humanTRB",
                                    "type": "OLGA"
                                }
                            },
                            "neg_reps": {
                                "signals": {},
                                "number_of_examples": 50,
                                "receptors_in_repertoire_count": 10,
                                "generative_model": {
                                    "default_model_name": "humanTRB",
                                    "type": "OLGA"
                                }
                            }
                        }
                    }
                }
            },
            "instructions": {
                "simulation_instruction": {
                    "type": "LigoSim",
                    "export_p_gens": False,
                    "max_iterations": 100,
                    "sequence_batch_size": 2000,
                    "simulation": "quickstart_simulation"
                }}
        }

        specs_file = path / "simulation_specs.yaml"
        with specs_file.open("w") as file:
            yaml.dump(specs, file)

        app = ImmuneMLApp(specs_file, path / "result")
        app.run()

        print("immuneML quickstart: finished generating a synthetic dataset.")

    def run(self, result_path: str):

        result_path = self.build_path(result_path)

        logging.basicConfig(filename=Path(result_path) / "log.txt", level=logging.ERROR, format='%(asctime)s %(levelname)s: %(message)s')
        warnings.showwarning = lambda message, category, filename, lineno, file=None, line=None: logging.warning(message)

        self._simulate_dataset_with_signals(result_path / "synthetic_dataset")

        print("immuneML quickstart: training a machine learning model...")
        specs_file = self.create_specfication(result_path / "machine_learning_analysis")
        app = ImmuneMLApp(specs_file, result_path / "machine_learning_analysis/result")
        app.run()

        print("immuneML quickstart: finished training a machine learning model.")


def main():
    quickstart = Quickstart()
    quickstart.run(sys.argv[1] if len(sys.argv) == 2 else None)


if __name__ == "__main__":
    main()
